;; sedit.el
;;
;; For the sMartedit project.  07/06/00
;; Dutch Meyer - dmeyer@u.washington.edu
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;########################### global variables #############################

(defvar smedit-debug-flag nil
  "*t if debugging is on, nil if debugging is off.")
(defvar smedit-move-delay 10
  "*number of ms to wait on repeated move commands, to slow interactive mode.")

(defvar smedit-debug-buffer nil 
  "Debugging information is sent to this buffer.")

(defvar smedit-debug-buffer-name "SmartEdit-Debug")
(defvar smedit-process-name "SmartEdit")

(defvar smedit-insert-commands (list "self-insert-command" "insert-buffer"
			      "newline" "split-line" "insert"
			      "insert-before-markers" "insert-char"
			      "insert-buffer-substring")
  "A list of all commands considered to be insertions.")

(defvar smedit-delete-commands (list "delete-backward-char" "erase-buffer" 
			      "delete-region" "delete-char"
			      "backward-delete-char-untabify"
			      "delete-horizontal-space"
			      "delete-indentation" "fixup-whitespace"
			      "just-one-space" "delete-blank-lines")
  "A list of all commands considered to be deletes.")

(defvar smedit-move-commands (list "next-line" "previous-line" "backward-char" 
			    "forward-char" "goto-char" "forward-word"
			    "backward-word" "beginning-of-buffer" 
			    "end-of-buffer"
			    "goto-line" "beginning-of-line" "end-of-line"
			    "forward-line" "vertical-motion" 
			    "move-to-window-line"
			    "forward-list" "backward-list" "up-list" 
			    "down-list"
			    "forward-sexp" "backawrd-sexp" "beginning-of-defun"
			    "end-of-defun" "skip-chars-forward" 
			    "skip-chars-backward"
			    "isearch-backward" "isearch-backward-regexp"
			    "isearch-cancel" "isearch-complete"
			    "isearch-forward" "isearch-forward-regexp"
			    "isearch-whitespace-chars")
  "A list of all commands considered to be moves.")

(defvar smedit-copy-commands (list "copy-region-as-kill" "kill-ring-save"
				   "clipboard-kill-ring-save"
				   "mouse-kill-ring-save")
  "A list of all commands considered to be copys.")

(defvar smedit-cut-commands (list "backward-kill-paragraph" 
				  "backward-kill-sentence"
			   "backward-kill-sexp" "backward-kill-word"
			   "clipboard-kill-region" "clipboard-kill-ring-save"
			   "isearch-yank-kill" "kill-comment" "kill-paragraph"
			   "kill-line" "kill-sexp" "kill-sentence" "kill-word")
  "A list of all commands considered to be cuts.")

(defvar smedit-paste-commands (list "yank" "yank-pop" "yank-rectangle"
			     "clipboard-yank")
  "A list of all commands considered to be pastes.")

(defvar smedit-cutSelection-commands (list "kill-region")
  "A list of all commands considered to cut from the mark to point.")

(defvar smedit-instant-failure-commands (list "append-next-kill" "undo"
		      "kill-rectangle" )
  "A list of commands we have no intention of modeling.")

(defvar smedit-ignore-commands (list 
		      "isearch-backward" "isearch-backward-regexp"
		      "isearch-cancel" "isearch-complete"
		      "isearch-forward" "isearch-forward-regexp"
		      "isearch-whitespace-chars")
  "These use the minibuffer, so we need to execute and let them run.")

;; The rest of these (while also global) will be make buffer local.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar smedit-keymap-mode nil
  "the SmartEdit keymap mode.")

(defvar smedit-editing-buffer nil
  "The buffer being edited with smartedit.")

(defvar smedit-process nil 
  "the SmartEdit process.")

(defvar smedit-playing-back nil 
  "t if SmartEdit is playing back, nil otherwise.")

(defvar smedit-event-type nil
  "The type of event that is currently being recorded.") 

(defvar smedit-undo-state nil 
  "The state the buffer was in.")
(defvar smedit-undo-point nil 
  "The location of the cursor.")
(defvar smedit-undo-cboard nil
  "The contents of the clipboard.")
(defvar smedit-undo-mark nil
  "The location of the mark.")
(defvar smedit-msg-str ""
  "Stores the string recieved from the Smartedit process.")
(defvar smedit-updated-hypothesis 0
  "0 - no update.  1 - update.  2 - update causing termination.")
(defvar smedit-overlay nil
  "The overlay used to highlight changes.")
(defvar smedit-kbd-macro nil
  "stores a normal macro as backup.")
(defvar smedit-num-steps 0
  "The number of steps we have left to complete.")
(defvar smedit-tolerance 100
  "We auto-accept any hypothesis over this tolerance.")
(defvar smedit-pretolerance 100
  "a holder for tolerance changes, so they don't conflict with the guess.")
(defvar smedit-cmd-array nil
  "The array of commands for minibuffer completion")
;;############################ User control functions ########################

(defun smedit-start-recording ()
  (interactive)
    (if (not smedit-process)
      (progn
	(setq smedit-process
	      (start-process smedit-process-name
			     nil
			     (concat smemacs-path "smemacs.py")))
	(set-process-filter smedit-process 'smedit-filter)
	(process-kill-without-query smedit-process)))

  (setq smedit-cmdarray nil)
  (mapatoms 'smedit-copy-command-to-array obarray)

  (if defining-kbd-macro
      (progn
	(message "While I appreciate your enthusiasm, I'm already recording."))
    (progn
      (setq old-echo echo-keystrokes)
      (setq echo-keystrokes 0)
      (start-kbd-macro nil)
      (setq smedit-event-type nil)
      (message "Hi, I'm SMARTedit, and I'm listening.")
      (if smedit-playing-back
	  (smedit-restore-undo-state))
      (let ((cmd nil))
	(while defining-kbd-macro
	 
	  (setq cmd (vconcat cmd 
			     (vector 
			      (smedit-macro-symbol-to-ascii 
			       (read-event)))))

	  (let ((translated (smedit-find-key cmd)))
	    (if translated
		(progn
		  (if (equal translated 'execute-extended-command)
		      (setq translated 
			    (intern (completing-read "M-x "
						     smedit-cmdarray
						     nil t))))
		  (if (eq (current-buffer) smedit-editing-buffer)
		      (progn
			(let ((new-event-type 
			       (smedit-get-command-type 
				(symbol-name translated))))
			  (if (equal new-event-type 'failure)
			      (progn
				(end-kbd-macro)
				(message (concat
					  (symbol-name translated)
					  " cannot be represented at"
					  " this time.  Exiting record."))
				(setq translated nil)))
			  (if (or (and (not 
					(equal smedit-event-type 
					       new-event-type))
				       (not (equal 'unknown new-event-type))
				       (not (equal 'failure new-event-type))
				       (not (and 
					     (equal smedit-event-type 'insert)
					     (equal new-event-type 'delete))))
				  (equal new-event-type 'cut-selection))
			      (progn
				(setq smedit-event-type new-event-type)
				(smedit-export-state))))))
		  (if (equal translated 'self-insert-command)
		      (setq translated cmd))
		  (if (and (symbolp translated)
			   (member (symbol-name translated)
			      smedit-ignore-commands))
		      (progn
			(funcall translated))
		    (if translated
			(progn
			  (command-execute translated))))
		  (setq cmd nil)))))))))

(defun smedit-guess-again ()
  "Display the next most likely hypothesis."
  (interactive)

  (if (not defining-kbd-macro)
      (if smedit-playing-back 
	  (progn
	    (smedit-restore-undo-state)
	    (smedit-export-next-guess))

	(message 
	 "Don't tell me I'm wrong before listening, try smedit-step-through"))
    (message "Maybe finish the recording, THEN start guessing, eh?")))

(defun smedit-step-through ()
  "Accepts the current hypothesis and moves on."
  (interactive)
  (if smedit-process
      (if defining-kbd-macro
	  (message "Please end your recording before stepping, thank you.")
	(progn
	  (smedit-accept-hypothesis)
	  (if smedit-playing-back
	      (smedit-guess-again))
	  (smedit-set-undo-state nil)))
    (message "I don't have a macro, perhaps there was a recording problem.")))

(defun smedit-end-recording ()
  (interactive)
  (if defining-kbd-macro
      (progn
	(setq echo-keystrokes old-echo)
	(setq smedit-kbd-macro last-kbd-macro)
	(setq smedit-event-type nil)
	(end-kbd-macro)
	(smedit-export-stop-recording)
	(message "I think I understand, and I have some suggestions."))
    (message "One cannot stop recording, if one never starts.")))

(defun smedit-exec-emacs-macro ()
  (interactive)
  (if smedit-kbd-macro
      (progn
	(message "You waste my intelligence on a simple emacs macro!?")
	(execute-kbd-macro smedit-kbd-macro))
    (message "I don't have an emacs macro in mind.")))

(defun smedit-reset ()
  (interactive)
  (smedit-silent-reset)
  (message "Good idea, let's start with a clean slate."))

(defun smedit-quick-step ()
  "A single key interface to the step/guess process."
  (interactive)

;  (smedit-restore-undo-state)
  (smedit-accept-hypothesis)
  (if (not (equal smedit-event-type 'terminate))
      (map-y-or-n-p 'smedit-prompter 'smedit-actor 'smedit-list
		    '("hypothesis" "guesses" "accept") 
		    '((?m smedit-excecute-n-steps 
			  "make n steps")
		      (?t smedit-set-tolerance 
			  "step until the guess is below some threshold")))))

(defun smedit-list ()
  "Makes a guess as to the next action."
  (progn
    (if (equal smedit-updated-hypothesis 2)
	(progn 
	  (smedit-silent-reset)
	  (not t))
      (progn
	(if (equal smedit-tolerance smedit-pretolerance)
	    (progn
	      (message "Forming hypothesis...")
	      (setq smedit-updated-hypothesis 0)
	      (smedit-guess-again)
	      (while (equal smedit-updated-hypothesis 0)
		(accept-process-output))))
	smedit-msg-str))))

(defun smedit-actor (obj)
  "Accepts the current hypothesis."
  (if (equal smedit-event-type 'delete)
      (delete-region 
       (overlay-start smedit-overlay)
       (overlay-end smedit-overlay)))
  (if (equal smedit-event-type 'terminate)
      (setq smedit-updated-hypothesis 2)
    (smedit-accept-hypothesis)))

(defun smedit-prompter (the-string)
  "prompts the formatted hypothesis."
  (progn
    smedit-msg-str))

(defun smedit-execute-n-steps ( &optional obj)
  "Excecutes some given number of steps."
  (interactive)
  (if smedit-process
      (if defining-kbd-macro
	  (message "Please end your recording before stepping, thank you.")
	(progn
	  (setq smedit-num-steps 
		(- (string-to-number 
		    (read-from-minibuffer 
		     "How many steps should I execute?")) 1))
	  (smedit-step-through)))
    (message "I don't have a macro, perhaps there was a recording problem.")))

(defun smedit-set-tolerance ( &optional obj)
  "Sets the tolerance level, under which SMARTedit will autoaccept."
  (interactive)
  (setq smedit-pretolerance 
	(string-to-number 
	 (read-from-minibuffer "What should the new tolerance be?"))))

(defun smedit-write-program (filename)
  (interactive "FEnter the filename for the program: ")
  (smedit-export-write-program filename))

(defun smedit-read-program (filename)
  (interactive "fEnter the filename for the program: ")
  (smedit-export-read-program filename))

(defun smedit-take-me-home ()
  (interactive)
  (message "you can never go home, but you can shop there."))

(defun smedit-count-me ()
  (interactive)
  (message (number-to-string num-input-keys)))
;;####################### Command oriented kinda things ####################

(defun smedit-silent-reset ()
  "resets SmartEdit back to a pre-recording state."
  (if defining-kbd-macro
      (smedit-end-recording))
  (setq smedit-kbd-macro nil)
  (setq smedit-playing-back nil)
  (setq smedit-event-type nil)
  (setq smedit-updated-hypothesis 0)
  (setq smedit-tolerance 100)
  (setq smedit-pretolerance 100)
  (smedit-set-undo-state nil)
  (setq smedit-msg-str "")
  (overlay-put smedit-overlay 'face nil)
  (if smedit-process
      (progn
	(smedit-export-shutup-and-die)
	(setq smedit-process nil))))

(defun smedit-accept-hypothesis ()
  "Accepts the current hypothesis."
  (if (equal smedit-event-type 'delete)
      (delete-region 
       (overlay-start smedit-overlay)
       (overlay-end smedit-overlay)))
  (setq smedit-tolerance smedit-pretolerance)
  (setq smedit-playing-back t)
  (smedit-set-undo-state nil)
  (if (equal smedit-event-type 'terminate)
      (smedit-silent-reset)
    (smedit-export-step-through)))

;;################### Recieving output from SMedit  #########################

(defun smedit-filter (proc string)
  "Handles the information coming from the smartedit process."

  (if smedit-debug-flag
      ;;Dumps the data out to the buffer, for good measure.
      (let ((old-buffer (current-buffer)))
	(unwind-protect
	    (let (moving)
	      (set-buffer smedit-debug-buffer)
	      (save-excursion
		(insert string)))
	  (set-buffer old-buffer))))
    
  ;; There can be only one hypothesis per message.
  (if (not (equal (string-match ":1:" string 0) nil))
      (smedit-handle-hypothesis 
       (substring string
		  (+ 3 (string-match ":1:" string 0)) nil))

    ;; We want to strip off any Errors.
    (if (not (equal (string-match ":2:" string 0) nil))
	(smedit-handle-error
	 (substring string
		    (+ 3 (string-match ":2:" string 0)) -1)))))


(defun smedit-handle-error (str)
  "Deals with an error by messaging it to the minibuffer."
  (setq smedit-event-type 'terminate)
  (setq smedit-updated-hypotheis 1)
  (message str))

(defun smedit-handle-hypothesis (str)
  "Handles a hypothesis made by smartedit."
       
  (cond
   ;; 0 -> terminate
   ((equal (substring str 0 1) "0")
    (let ((str (substring str 2 nil)))
      (smedit-set-undo-state (buffer-substring-no-properties 
			      (point-min) (point-max)) 
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str)))
	    (str (smedit-strip-next-cmd str)))
	(setq smedit-msg-str 
	      (concat guess ". " (smedit-get-next-cmd str)))))
    (setq smedit-updated-hypothesis 1)
    (setq smedit-event-type 'terminate)
    (message (smedit-prep smedit-msg-str)))
   
   ;; 1 -> insert
   ((equal (substring str 0 1) "1") 
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'insert)
      (smedit-set-undo-state (buffer-substring-no-properties 
			      (point-min) (point-max)) 
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (let ((start-point (point)))
	(insert (smedit-fix-special-chars 
		 (smedit-fix-octal (substring 
			     (smedit-get-next-cmd str) 1 -1))))
	(overlay-put smedit-overlay 'face 'secondary-selection)
	(move-overlay smedit-overlay start-point (point)))
      (let ((str (smedit-strip-next-cmd str)))
	(let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	  (let ((str (smedit-strip-next-cmd str)))
	    (setq smedit-msg-str 
		  (concat guess ". " (smedit-get-next-cmd str)))
	    (if ( < smedit-tolerance 
		    (string-to-number 
		     (substring (smedit-get-next-cmd str) 1 -8)))
		(progn
		  (sit-for 0 smedit-move-delay)
		  (smedit-step-through)))))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 0)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))
    
   
   ;; 2 -> move
   ((equal (substring str 0 1) "2")
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'move)
      (smedit-set-undo-state (buffer-substring-no-properties
			      (point-min) (point-max)) 
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (set-window-point 
       (get-buffer-window (current-buffer)) 
       (+ 1 (string-to-number (smedit-get-next-cmd str))))
      (overlay-put smedit-overlay 'face 'secondary-selection)
      (move-overlay smedit-overlay
		    (+ 1 (string-to-number (smedit-get-next-cmd str)))
		    (+ 2 (string-to-number (smedit-get-next-cmd str))))
      (let ((str (smedit-strip-next-cmd str)))
	(let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	  (let ((str (smedit-strip-next-cmd str)))
	    (setq smedit-msg-str 
		  (concat guess ". " (smedit-get-next-cmd str)))
	    (if ( < smedit-tolerance 
		    (string-to-number 
		     (substring (smedit-get-next-cmd str) 1 -8)))
		(smedit-step-through))))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 0)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))

   ;; 3 -> delete
   ((equal (substring str 0 1) "3") 
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'delete)
      (smedit-set-undo-state (buffer-substring-no-properties 
			      (point-min) (point-max))
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (overlay-put smedit-overlay 'face 'secondary-selection)
      (move-overlay 
       smedit-overlay 
       (+ (string-to-number (smedit-get-next-cmd str)) 1)
       (+ (string-to-number (smedit-get-next-cmd 
			     (smedit-strip-next-cmd str))) 1))
      (let ((str (smedit-strip-next-cmd (smedit-strip-next-cmd str))))
	(let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	  (let ((str (smedit-strip-next-cmd str)))
	    (setq smedit-msg-str 
		  (concat guess ". " (smedit-get-next-cmd str)))
	    (if ( < smedit-tolerance (string-to-number 
				      (substring 
				       (smedit-get-next-cmd str) 1 -8)))
		(smedit-step-through))))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 0)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))

   ;; 4 -> cut
   ((equal (substring str 0 1) "4") 
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'cut)
      (smedit-set-undo-state (buffer-substring-no-properties
			      (point-min) (point-max))
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (kill-region (mark) (point))
      (let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	(let ((str (smedit-strip-next-cmd str)))
	  (setq smedit-msg-str 
		(concat guess ". " (smedit-get-next-cmd str)))
	  (if ( < smedit-tolerance 
		  (string-to-number 
		   (substring (smedit-get-next-cmd str) 1 -8)))
	      (smedit-step-through)))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 0)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))

   ;; 5 -> copy
   ((equal (substring str 0 1) "5") 
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'copy)
      (smedit-set-undo-state (buffer-substring-no-properties
			      (point-min) (point-max)) 
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (copy-region-as-kill (mark) (point))
      (let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	(let ((str (smedit-strip-next-cmd str)))
	  (setq smedit-msg-str 
		(concat guess ". " (smedit-get-next-cmd str)))
	  (if ( < smedit-tolerance 
		  (string-to-number 
		   (substring (smedit-get-next-cmd str) 1 -8)))
	      (smedit-step-through)))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 0)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))

   ;; 6 -> paste
   ((equal (substring str 0 1) "6") 
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'paste)
      (smedit-set-undo-state (buffer-substring-no-properties
			      (point-min) (point-max))
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (let ((start-point (point)))
	(insert (car kill-ring-yank-pointer))
	(overlay-put smedit-overlay 'face 'secondary-selection)
	(move-overlay smedit-overlay start-point (point)))      
      (let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	(let ((str (smedit-strip-next-cmd str)))
	  (setq smedit-msg-str 
		(concat guess ". " (smedit-get-next-cmd str)))
	    (if ( < smedit-tolerance 
		    (string-to-number 
		     (substring (smedit-get-next-cmd str) 1 -8)))
		(smedit-step-through)))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 0)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))

   ;; 7 -> select
   ((equal (substring str 0 1) "7")
    (let ((str (substring str 2 nil)))
      (setq smedit-event-type 'select)
      (smedit-set-undo-state (buffer-substring-no-properties
			      (point-min) (point-max)) 
			     (point)
			     (mark)
			     (car kill-ring-yank-pointer))
      (set-mark (+ 1 (string-to-number (smedit-get-next-cmd str))))
      (overlay-put smedit-overlay 'face 'secondary-selection)
      (move-overlay smedit-overlay (mark) (point))
      (let ((str (smedit-strip-next-cmd str)))
	(let ((guess (smedit-fix-special-chars (smedit-get-next-cmd str))))
	  (let ((str (smedit-strip-next-cmd str)))
	    (setq smedit-msg-str 
		  (concat guess ". " (smedit-get-next-cmd str)))
	    (if ( < smedit-tolerance 
		    (string-to-number 
		     (substring (smedit-get-next-cmd str) 1 -8)))
		(smedit-step-through))))))
    (setq smedit-updated-hypothesis 1)
    (message (smedit-prep smedit-msg-str))
    (if ( > smedit-num-steps 1)
	(progn
	  (setq smedit-num-steps (- smedit-num-steps 1))
	  (smedit-step-through))))))

;;####################### Initialization functions #########################

(defun smedit-remove-from-buffer()
  "Removes smartedit from the current buffer."
  (interactive)
  (progn

    (smedit-silent-reset)

    (delete-overlay smedit-overlay)
    
    (if smedit-debug-flag
	(kill-buffer 'smedit-debug-buffer))

    (define-key smedit-keymap-mode  [menu-bar smedit]
     'undefined)

    (kill-local-variable 'smedit-debug-buffer)
;;    (kill-local-variable 'smedit-keymap-mode)
    (setq smedit-keymap-mode nil) ;; this instead.
    (kill-local-variable 'smedit-editing-buffer)
    (kill-local-variable 'smedit-process)
    (kill-local-variable 'smedit-playing-back)
    (kill-local-variable 'smedit-event-type)
    (kill-local-variable 'smedit-undo-state)
    (kill-local-variable 'smedit-undo-point)
    (kill-local-variable 'smedit-undo-cboard)
    (kill-local-variable 'smedit-undo-mark)
    (kill-local-variable 'smedit-msg-str)
    (kill-local-variable 'smedit-updated-hypothesis)
    (kill-local-variable 'smedit-overlay)
    (kill-local-variable 'smedit-kbd-macro)
    (kill-local-variable 'smedit-tolerance)
    (kill-local-variable 'smedit-pretolerance)
    (kill-local-variable 'smedit-cmd-array)))

(defun smedit-start-up()
  "Makes the necessary bindings to start smartedit for any buffer."
  (progn

    (make-local-variable 'smedit-debug-buffer)
    (make-local-variable 'smedit-keymap-mode)
    (make-local-variable 'smedit-editing-buffer)
    (make-local-variable 'smedit-process)
    (make-local-variable 'smedit-playing-back)
    (make-local-variable 'smedit-event-type)
    (make-local-variable 'smedit-undo-state)
    (make-local-variable 'smedit-undo-point)
    (make-local-variable 'smedit-undo-cboard)
    (make-local-variable 'smedit-undo-mark)
    (make-local-variable 'smedit-msg-str)
    (make-local-variable 'smedit-updated-hypothesis)
    (make-local-variable 'smedit-overlay)
    (make-local-variable 'smedit-kbd-macro)
    (make-local-variable 'smedit-tolerance)
    (make-local-variable 'smedit-pretolerance)
    (make-local-variable 'smedit-cmd-array)

    (setq smedit-editing-buffer (current-buffer))
    (setq smedit-overlay (make-overlay 0 0))
    (overlay-put smedit-overlay 'face nil)


    ;; Make the standard key-bindings.
    (progn
      (setq smedit-keymap-mode (make-sparse-keymap))
      (define-key smedit-keymap-mode [ ?\C-x ?( ] 
	'smedit-start-recording)
      (define-key smedit-keymap-mode [ ?\C-x ?) ]                    
      'smedit-end-recording)
      (define-key smedit-keymap-mode [ ?\C-c ?g ] 
	'smedit-guess-again)
      (define-key smedit-keymap-mode [ ?\C-x ?e ] 
	'smedit-step-through )
      (define-key smedit-keymap-mode [ ?\C-c ?r ] 
	'smedit-reset )
      (define-key smedit-keymap-mode [ ?\C-c ?u ] 
	'smedit-restore-undo-state)
      (define-key smedit-keymap-mode [ ?\C-c ?n ] 
	'smedit-execute-n-steps)
      (define-key smedit-keymap-mode [ ?\C-c ?t ] 
	'smedit-set-tolerance)
      (define-key smedit-keymap-mode [ ?\C-c ?q ] 
	'smedit-quick-step )
      (define-key smedit-keymap-mode [ ?\C-c ?m ]
	'smedit-exec-emacs-macro )
      (define-key smedit-keymap-mode  [menu-bar smedit]
	(cons "SMARTedit" (make-sparse-keymap "SMARTedit")))
      (define-key smedit-keymap-mode
	[menu-bar smedit record]
	'("Start Recording" . smedit-start-recording))
      (define-key smedit-keymap-mode
	[menu-bar smedit stop]
	'("End Recording" . smedit-end-recording))
      (define-key smedit-keymap-mode
	[menu-bar smedit reset]
	'("Reset SMARTedit" . smedit-reset))
      (define-key smedit-keymap-mode
	[menu-bar smedit normal]
	'("Exec Normal Macro" . smedit-exec-emacs-macro))
      (define-key smedit-keymap-mode
	[menu-bar smedit step]
	'("Make Step" . smedit-step-through))
      (define-key smedit-keymap-mode
	[menu-bar smedit guess]
	'("Guess Again" . smedit-guess-again))
      (define-key smedit-keymap-mode
	[menu-bar smedit quick]
	'("Quick-Step" . smedit-quick-step))
      (define-key smedit-keymap-mode
	[menu-bar smedit stepn]
	'("Make n Steps" . smedit-execute-n-steps))
      (define-key smedit-keymap-mode
	[menu-bar smedit tol]
	'("Set Tolerance" . smedit-set-tolerance))
      (use-local-map smedit-keymap-mode))
    
    (modify-frame-parameters (selected-frame) '((menu-bar-lines . 2)))
	;; Add the SmartEdit mode to the minor-mode-alist
	(if (not (assq 'smedit-keymap-mode minor-mode-alist))
	    (setq minor-mode-alist
		  (cons '(smedit-keymap-mode " SMARTedit")
		        minor-mode-alist)))

    (if smedit-debug-flag
	;; Create the SmartEdit buffer
	(setq smedit-debug-buffer (get-buffer
			    (generate-new-buffer smedit-debug-buffer-name))))))

;;######################### Helper functions ################################

(defun smedit-copy-command-to-array (arg)
  "if the arg is a command, copy it to smedit-cmdarray"
  (if (commandp arg)
	(setq smedit-cmdarray 
	      (cons 
	       (cons (symbol-name arg) nil) 
	       smedit-cmdarray))))

(defun smedit-set-undo-state (string &optional loc sel clip)
  "Stores the current buffer state."
  (progn
    (setq smedit-undo-state string)
    (setq smedit-undo-point loc)
    (setq smedit-undo-mark sel)
    (setq smedit-undo-cboard clip)))

(defun smedit-restore-undo-state ()
  "Restores a previously stored buffer state."
  (interactive)
  (if (or defining-kbd-macro smedit-playing-back )
      (smedit-export-restore-state))
  (if smedit-undo-state
      (progn
	(overlay-put smedit-overlay 'face nil)
	(erase-buffer)
	(insert smedit-undo-state)
	(if smedit-undo-point
	    (set-window-point (get-buffer-window (current-buffer)) 
			      smedit-undo-point))
	(if smedit-undo-mark
	    (set-mark smedit-undo-mark))
	(if smedit-undo-cboard
	    (kill-new smedit-undo-cboard)))))


(defun smedit-macro-symbol-to-ascii (item)
  (cond ;;((equal item 'cancelBAD) 7)
	((equal item 'tab) 9)
	;;((equal item 'newlineBAD) 10)
	;;((equal item 'vertabBAD) 11)
	;;((equal item 'formfeedBAD) 12)
	((equal item 'return) 13)
	((equal item 'escape) 27)
	((equal item 'delete) 127)
	(t item)))

(defun smedit-find-key (command)
  "If there's a key in the global or local map, this will find it, else nil."
  (let ((key (lookup-key (current-local-map) command)))
    (if (or (numberp key)
	    (keymapp key)
	    (equal nil key))
	(let ((key (lookup-key (current-global-map) command)))
	  (if (or (numberp key)
		  (keymapp key))
	      nil
	    key))
      key)))
	
(defun smedit-get-command-type (command)
  "Classifies a command according to the smartedit categories."
  (cond ((member command smedit-insert-commands) 'insert)
        ((member command smedit-move-commands) 'move)
        ((member command smedit-delete-commands) 'delete)
        ((member command smedit-copy-commands) 'copy)
        ((member command smedit-paste-commands) 'paste)
        ((member command smedit-cut-commands) 'cut)
	((member command smedit-cutSelection-commands) 'cutSelection)
	((member command smedit-instant-failure-commands) 'failure)
        (t 'unknown)))

(defun smedit-get-text-buffer-str ()
  "Returns a formated string of the text in the buffer."
  (format "%s :" 
	  (smedit-escape-special-chars
	   (buffer-substring-no-properties (point-min) (point-max)))))

(defun smedit-get-cursor-loc-str ()
  "Returns a formated string of the cursor location."
  (format "%s :" (- (point) 1)))

(defun smedit-get-clipboard-str ()
  "Returns a formated string of the first position on the kill ring contents."
  (format "%s :" (if (not (equal kill-ring-yank-pointer nil))
		     (smedit-escape-special-chars 
		      (car kill-ring-yank-pointer)) "")))

(defun smedit-get-selection-str ()
  "Returns a formated string of the selection extent."
  (let ((mark-val (if (mark)
		      (mark)
		    (point))))
    (concat (number-to-string (- (point) 1 )) ", " 
	    (number-to-string (- mark-val 1)) " :" )))

(defun smedit-get-current-event-str ()
  "Returns a formated string of the current event type."
  (format "%s :" smedit-event-type))

;; Returns the state of the buffer as a string of 5 fields separated by " :"
;; and ending with an additional line feed.  The fields are as follows:
;;  <0> The current event type
;;  <1> The cursor location
;;  <2> The selection extent
;;  <3> The clipboard
;;  <4> The text buffer
(defun smedit-get-status-str ()
  "Returns the buffer state as a formated string."
  (concat (smedit-get-current-event-str) 
	  (smedit-get-cursor-loc-str) (smedit-get-selection-str) 
	  (smedit-get-clipboard-str) (smedit-get-text-buffer-str) "\n" ))

(defun smedit-get-end-str ()
  "Returns the buffer state, and indicates the end of a recording."
  (concat "END_RECORDING :" (smedit-get-cursor-loc-str) 
	  (smedit-get-selection-str)
	  (smedit-get-clipboard-str) (smedit-get-text-buffer-str) "\n" ))

(defun smedit-get-step-str ()
  "Returns the buffer state, with a command to step."
  (concat "STEP :" (smedit-get-cursor-loc-str) (smedit-get-selection-str) 
	  (smedit-get-clipboard-str) (smedit-get-text-buffer-str) "\n"))

(defun smedit-escape-special-chars (str)
  "Escapes '\' and ':' characters with a backslash."
  (progn

    ;; mystring will be the return result.
    (setq mystring str)

    ;; escape the backslashes first.
    (setq loc -1)
    (while (not (equal loc nil))
      (progn
	(setq loc (+ 2 loc))
	(setq loc (string-match "\\\\" mystring loc))
	(if (not (equal loc nil))
	    (setq mystring (replace-match "\\\\\\\\" nil nil mystring nil)))))
    
    ;; then escape ':'
    (setq loc -1)
    (while (not (equal loc nil))
      (progn
	(setq loc (+ 2 loc))
	(setq loc (string-match ":" mystring loc))
	(if (not (equal loc nil))
	    (setq mystring (replace-match "\\\\:" nil nil mystring nil))))))
  
  ;; Return the resulting string (better way to do this??)
  mystring)


(defun smedit-get-next-cmd (str)
  "Gets the next field of a message from smartedit."
  (progn
    (setq loc (string-match ":" str 0))
    (while (equal "\\" (substring str (- loc 1) loc))
      (setq loc (string-match ":" str (+ 1 loc))))
    (substring str 0 loc)))

(defun smedit-fix-special-chars (foo)
  "Removes the extra backslash escaping '\' and ':'."
  (progn
    (setq mystring "")
    (setq locB -1)
    (while locB
      (progn
	(setq locA (+ 1 locB))
	(setq locB (string-match "\\\\" foo locA))
	(if locB
	    (setq locB (+ 1 locB)))
	(setq mystring (concat mystring (substring foo locA locB)))))
    mystring))

(defun smedit-fix-octal (foo)
  (progn
    (setq mystring "")
    (setq locB -1)
    (while locB
      (progn
	(setq locA (+ locB 1))
	(setq locB (string-match "\\0" foo locA))
	(if locB
	    (progn
	      (setq cool (substring foo (- locB 1) (+ locB 3)))
	      (if (equal cool "\\012")
		  (setq cool "\n")
		(if (equal cool "\\011")
		    (setq cool "\t")))
	      (setq mystring
		    (concat mystring
			    (substring foo locA (- locB 1))
			    (char-to-string (string-to-char cool))))
	      (setq locB (+ locB 2)))
	  (setq mystring
		(concat mystring
			(substring foo locA nil))))))
    mystring))

(defun smedit-strip-next-cmd (str)
  "Removes the next field of a message from smartedit."
  (progn
    (setq loc (string-match ":" str 0))
    (while (equal "\\" (substring str (- loc 1) loc))
      (setq loc (string-match ":" str (+ 1 loc))))

    (substring str (+ 1 loc) nil)))

(defun smedit-prep (str)
  "Prepares a string to be messaged to the minibuffer."
  (progn
    
    ;; mystring will be the return result.
    (setq mystring str)
    
    (setq loc -1)
    (while (not (equal loc nil))
      (progn
	(setq loc (+ 2 loc))
	(setq loc (string-match "%" mystring loc))
	(if (not (equal loc nil))
	 (setq mystring (replace-match "%%" nil nil mystring nil)))))
    mystring))

;; ############################ Messaging ###############################

(defun smedit-export-restore-state ()
  "Sends a command to restore to smartedit."
  (process-send-string smedit-process "RESTORE :\n")
  (process-send-eof smedit-process))

(defun smedit-export-state ()
  "Sends the state of the buffer to smartedit."
  (process-send-string smedit-process (smedit-get-status-str))
  (process-send-eof smedit-process))

(defun smedit-export-stop-recording ()
  "Sends a command to end the current recording to smartedit."
  (process-send-string smedit-process (smedit-get-end-str))
  (process-send-eof smedit-process))

(defun smedit-export-reset ()
  "Sends the command to reset smartedit."
  (process-send-string smedit-process "RESET :\n")
  (process-send-eof smedit-process))

(defun smedit-export-next-guess ()
  "Sends the command to make another guess to smartedit."
  (process-send-string smedit-process "NEXT_GUESS :\n")
  (process-send-eof smedit-process))

(defun smedit-export-step-through ()
  "Sends the step-through command to smartedit."
  (process-send-string smedit-process (smedit-get-step-str))
  (process-send-eof smedit-process))

(defun smedit-export-shutup-and-die ()
  "Tells smartedit that it's not so smart after all, cause it's about to die."
  (process-send-string smedit-process "DIE_DIE_DIE :\n")
  (process-send-eof smedit-process))

(defun smedit-export-write-program (filename)
  "Sends the write command to save the program as the given filename."
  (process-send-string smedit-process 
		       (concat "SAVE_PROGRAM :" 
			       (expand-file-name filename) 
			       " :\n"))
  (process-send-eof smedit-process))

(defun smedit-export-read-program (filename)
  "Sends the write command to load a program from the given filename."
  (process-send-string smedit-process 
		       (concat "LOAD_PROGRAM :" 
			       (expand-file-name filename)
			       " :\n"))
  (process-send-eof smedit-process))

;; We want to run start-up just once, at load time.
(smedit-start-up)
