import pqueue
import math
import copy
import operator
import time
from threading import *

LG_2 = math.log(2)
def ent_log(x):
	if x:
		return math.log(x) / LG_2
	else:
		return 0
	
def calc_entropy(outcomes):
	# Handy general function for calculating entropy of a number of outcomes.
	# Each outcome should be a probability. All probabilities together
	# should sum to 1. If they don't (if they sum to less than 1) the
	# remaining is automatically added as another outcome.
	
	for i in range(len(outcomes)):
		if not (0.0 <= outcomes[i] <= 1.0):
			print 'Probability out of bounds in calc_entropy:', outcomes[i]
			if outcomes[i] < 0.0:
				outcomes[i] = 0.0
			else:
				outcomes[i] = 1.0
	total = reduce(operator.add, outcomes, 0)
	## I'm commenting this out because numerical instability makes it fire
	## at ridiculous times (when the total really should be <= 1.0).
	# assert(0.0 <= total <= 1.0)

	if total < 1:
		outcomes.append(1 - total)

	return reduce(operator.add,
			    map(lambda n, lg = ent_log: -n * lg(n), outcomes), 0)

##         # For efficiency and in order to find "small" examples,
##         # truncate the input state to a window in each direction.
##         WINDOW_PANE = 100
##         new_instate = copy.copy(instate)
##         left = clip(new_instate.pos - WINDOW_PANE)
##         right = clip(new_instate.pos + WINDOW_PANE)
##         new_instate.contents = instate.contents[left:right]
##         new_instate.pos = instate.pos - left
##         if new_instate.selection_extent:
##             new_instate.selection_extent = \
##                   (clip(new_instate.selection_extent[0] - left),
##                    clip(new_instate.selection_extent[1] - left))
##             if new_instate.selection_extent[0] == \
##                new_instate.selection_extent[1]:
##                 new_instate.selection_extent = None





def clip(num, left = 0, right = None):
	"""Clip num to the range [left, right].

	If right is None, right = infinity.

	Precondition: left <= right"""

	assert(right is None or left <= right)
	
	if right is None:
		right = num

	if left <= num <= right:
		return num
	elif num < left:
		return left
	else:
		return right


class EntropyCalculator:
	def __init__(self, log_leaf_counts, steps,
			   hypothesis_generator, simulator, instate):
		self.length = steps
		self.hg = hypothesis_generator
		self.simulator = simulator
		self.log_counts = log_leaf_counts
		# Add an extra entry for padding. i.e., no branch points off the
		# end of the program.
		self.log_counts.append(0)
		
		# The loose ends are unexpanded nodes in the entropy
		# calculation tree. Each one is characterized by its input
		# state, its step, and its probability. Its value is
		# recorded as its probability multiplied by the log of its
		# branching factor (the minimum improvement in error bounds
		# that would result from expanding it).

		self.loose_ends = pqueue.Heap()

		if self.length > 0:
			self.loose_ends.insert((instate, 0, 1.0),
							   -(self.log_counts[0] - \
								self.log_counts[1]))

		self.error = self.log_counts[0]
		self.entropy = 0

	def finished(self):
		return self.loose_ends.is_empty()

	def next_step_value(self):
		return -self.loose_ends.peek_min_value()

	def step(self):
		state = self.loose_ends.delete_min()

		instate = state[0]
		step = state[1]
		prob = state[2]

		# The entropy assuming no branch points under this node
		local_entropy = -prob * ent_log(prob)

		# The error at this node (diff between above and the entropy
		# if all possible branchings occur with equal probability under
		# this node).
		local_error = prob * self.log_counts[step]

		# Remove this node from the running entropy/error counts
		self.entropy = self.entropy - local_entropy
		self.error = self.error - local_error

		if step == self.length:
			raise 'overstepped length of forecast in entropy calculator'

		result_set = self.hg(instate, step)
		if not result_set:
			#DEBUG!!!
			print 'Stopping due to lack of hypotheses.'			
			# No more predictions
			self.entropy = self.entropy + local_entropy
			return

		result_set.simple_init()
		hypotheses = result_set.get_grouped_actions()
		if not hypotheses:
			#DEBUG!!!
			print 'Stopping due to wierd lack of hypotheses.'
			print 'PROBABLY A BUG!!' + '!'
			# No more predictions
			self.entropy = self.entropy + local_entropy
			return

		for (totalprob, hyplist) in hypotheses:
			totalprob = totalprob * prob

			hypothesis = hyplist[0][0].hyp
			outstate = self.simulator(hypothesis, instate)

			# No matter what, add in the optimistic entropy (as if this
			# were a branchless point).
			sub_local_entropy = -totalprob * ent_log(totalprob)
			self.entropy = self.entropy + sub_local_entropy

			# If there is an outstate, set up to recurse into it (and
			# account for possible error).
			if outstate:
				sub_local_error = totalprob * self.log_counts[step+1]
				self.error = self.error + sub_local_error
				if step + 1 < self.length:
					sub_local_value = totalprob * \
								   (self.log_counts[step+1] -
								    self.log_counts[step+2])
					self.loose_ends.insert((outstate,
									    step + 1,
									    totalprob),
									   -sub_local_value)
	
	def get_error(self):
		return self.error

	def get_entropy_estimate(self):
		return self.entropy
		
	def timed_approx_calc(self, time_limit = None,
					  tolerance_percent = None):
		# Establish the time limit
		stop_running = Event()
		stop_running.clear()
		if time_limit:
			def kill_run(limit = time_limit, flag = stop_running):
				time.sleep(limit)
				flag.set()
				
			Thread(target = kill_run).start()
		
		if tolerance_percent is None:
			tolerance_percent = 0

		# Change tolerance to a factor rather than a percent:
		tolerance = 1 + tolerance_percent

		while not stop_running.isSet() and not self.finished() and \
			 self.get_error() > tolerance * self.get_entropy_estimate():
			self.step()

def choose_biggest_incr_except_glb(prospects, glb, glb_choice,
							gub, gub_choice):
	best = None
	best_incr = -1
	for (choice, incr) in prospects.items():
		if incr > best_incr and choice != glb_choice:
			best = choice
			best_incr = incr

	if best is None:
		return choose_biggest_incr(prospects, glb, glb_choice,
							  gub, gub_choice)
	else:
		return best

def choose_gub(prospects, glb, glb_choice, gub, gub_choice):
	if prospects.has_key(gub_choice):
		return gub_choice
	else:
		return choose_biggest_incr(prospects, glb, glb_choice,
							  gub, gub_choice)

def choose_biggest_incr(prospects, glb, glb_choice,
				    gub, gub_choice):
	best = None
	best_incr = -1
	for (choice, incr) in prospects.items():
		if incr > best_incr:
			best = choice
			best_incr = incr

	return best

def choose_biggest_incr_of_glb_and_gub(prospects, glb, glb_choice,
							    gub, gub_choice):
	best = None
	best_incr = -1
	for (choice, incr) in prospects.items():
		if incr > best_incr and \
		   (choice == glb_choice or choice == gub_choice):
			best = choice
			best_incr = incr

	if best is not None:
		return best
	else:
		return choose_biggest_incr(prospects,glb,glb_choice,gub,gub_choice)

def choose_next_step(prospects, glb, glb_choice, gub, gub_choice):
	# Trying out some different ways to pick the next
	# calculator to ping.

	return choose_biggest_incr_of_glb_and_gub(prospects, glb, glb_choice,
									  gub, gub_choice)

	

def choose_highest_entropy(entropy_calculators, tolerance,
					  time_limit = 0.75,
					  step_limit = None):
	# time_limit of None means no limit.
	# step_limit of None means no step limit.
	glb_choice = None
	glb = 0
	gub = 0

	# Establish the time limit
	stop_running = Event()
	stop_running.clear()
	if time_limit:
		def kill_run(limit = time_limit, flag = stop_running):
			time.sleep(limit)
			flag.set()
		
		Thread(target = kill_run).start()

	# Change tolerance to a factor rather than a percent:
	tolerance = 1 + tolerance

	prospects = {}
	for calc in entropy_calculators:
		prospects[calc] = calc.next_step_value()

	# Initialize the bounds.
	upper_bounds = []
	for calc in entropy_calculators:
		est = calc.get_entropy_estimate()
		err = calc.get_error()

		if est >= glb:
			glb = est
			glb_choice = calc
		
		if est + err >= gub:
			gub = est + err
			gub_choice = calc

		upper_bounds.append(est+err)

	#DEBUG!!
	#print 'Initial bounds:'
	#print 'Greatest entropy estimate:', glb
	#print 'Greatest possible entropy:', gub
	#print 'Choice is:', entropy_calculators.index(glb_choice)

	steps = 0
	# Continue until the greatest upper bound (the highest any entropy
	# at any location could possibly be) is no more than TOLERANCE higher
	# than the greatest lower bound (the highest guaranteed entropy).
	while not stop_running.isSet() and steps != step_limit and \
		 gub > glb * tolerance and prospects != {}:
		if len(prospects) == 1 and prospects.keys()[0] == glb_choice:
			# Nothing more to improve but the correct answer.
			break

		next_calc = choose_next_step(prospects, glb, glb_choice,
							    gub, gub_choice)
		del prospects[next_calc]
		if next_calc is None:
			#DEBUG!!
			#print 'Breaking out b/c no good choice remains.'
			break

		# Only work with the calculator if it could possibly be our
		# eventual choice.
		if next_calc.get_entropy_estimate() + \
		   next_calc.get_error() <= glb * tolerance:
			#DEBUG!!
			#print 'Throwing away choice', \
			#	 entropy_calculators.index(next_calc), \
			#	 'due to too low possible entropy.'
			continue

		# OK, this might be my target, work with it for one more step.
		next_calc.step()

		this_lb = next_calc.get_entropy_estimate()
		this_ub = this_lb + next_calc.get_error()

		# Update the gub downward if we have sufficiently descreased
		# this calc's bound.
		# This should really use a pq with the ability to decreaseKey.
		if gub_choice == next_calc:
			gub = this_ub
			for calc in entropy_calculators:
				est = calc.get_entropy_estimate()
				err = calc.get_error()
				
				if est + err >= gub:
					gub = est + err
					gub_choice = calc
		
		# Reinsert the calculator if there is more work to be done
		# on this estimate.
		if not next_calc.finished():
			prospects[next_calc] = next_calc.next_step_value()
			
		if this_lb > glb:
			glb = this_lb
			glb_choice = next_calc
		
		steps = steps + 1
		#DEBUG!!!
		#print 'After', steps, 'steps, bounds are:'
		#print 'Greatest entropy estimate:', glb
		#print 'Greatest possible entropy:', gub
		#print 'Choice is:', entropy_calculators.index(glb_choice)

	#DEBUG!!!
	#print 'Best choice has entropy est.', glb_choice.get_entropy_estimate(),\
	#	 'and error', glb_choice.get_error()
	#print 'Highest possible is', gub
	#print 'Achieved in', steps, 'steps'

	return entropy_calculators.index(glb_choice)
	
