class PriorityQueue:
    def delete_min(self):
        """Deletes and returns the minimal element."""

        raise 'abstract'

    def peek_min(self):
        """Returns the minimal element."""

        raise 'abstract'

    def peek_min_value(self):
        """Returns the minimal element's value."""

        raise 'abstract'

    def insert(self, object, value = None):
        """Inserts the object into the queue with the given value.

        If value is None, value is set equal to object."""

        raise 'abstract'

    def size(self):
        """Return the number of elements in the queue."""

        raise 'abstract'

    def is_empty(self):
        """Returns 1 iff the queue is empty."""

        return self.size() == 0

    def values(self):
        """Returns a list of the values in the priority queue."""

        raise 'abstract'

class Heap(PriorityQueue):
    def __init__(self):
        self.pairs = []
        self.pairs_size = 0

    def delete_min(self):
        if self.pairs_size == 0:
            raise 'attempt to delete from empty heap'

        object = self.pairs[0][0]
        self.pairs_size = self.pairs_size - 1
        self.pairs[0] = self.pairs[self.pairs_size]
        del self.pairs[self.pairs_size]
        self.percolate_down()
        
        return object

    def peek_min(self):
        assert(self.pairs_size > 0)

        return self.pairs[0][0]

    def peek_min_value(self):
        assert(self.pairs_size > 0)

        return self.pairs[0][1]

    def insert(self, object, value = None):
        if value is None:
            value = object

        self.pairs.append((object, value))
        self.pairs_size = self.pairs_size + 1
        self.percolate_up()

    def size(self):
        return self.pairs_size

    def values(self):
        return map(lambda x: x[1], self.pairs)

    def parent(self, location):
        if location == 0:
            return None
        else:
            return (location - 1) / 2

    def left_child(self, location):
        child = location * 2 + 1
        if child >= self.pairs_size:
            return None
        else:
            return child

    def right_child(self, location):
        sibling = self.left_child(location)
        if sibling is None:
            return None

        child = sibling + 1
        if child >= self.pairs_size:
            return None
        else:
            return child

    def percolate_up(self, location = None):
        if location is None:
            location = self.pairs_size - 1

        parent = self.parent(location)

        if parent is None:
            # Already at the root, quit
            return
        else:
            # If the parent has a greater value, swap and continue.
            if self.pairs[parent][1] > self.pairs[location][1]:
                temp = self.pairs[parent]
                self.pairs[parent] = self.pairs[location]
                self.pairs[location] = temp
                location = parent
                self.percolate_up(location)

    def percolate_down(self, location = None):
        if location is None:
            location = 0

        left = self.left_child(location)
        right = self.right_child(location)
        if left is None:
            # Done; no further to percolate
            return
        elif right is None:
            # Special case: only one child
            if self.pairs[location][1] > self.pairs[left][1]:
                temp = self.pairs[left]
                self.pairs[left] = self.pairs[location]
                self.pairs[location] = temp
            return
        else:
            # Two children. Just the usual.

            # Find the lesser child
            if self.pairs[left][1] < self.pairs[right][1]:
                lesser = left
            else:
                lesser = right

            # If the key is greater than the lesser child, swap it.
            if self.pairs[location][1] > self.pairs[lesser][1]:
                temp = self.pairs[lesser]
                self.pairs[lesser] = self.pairs[location]
                self.pairs[location] = temp
                location = lesser
                self.percolate_down(location)
                
        
