#!/usr/bin/env python

# TODO:
# Resolve problems with actions that are only distinguishable BEFORE we
# perform them (like backspace/insert char as they interact with the
# selection)!!!
# keep track of cursor column as done in text.tcl
# detect table columns (make field # an attribute)
# replace pickle with cpickle and use binary mode

from Tkinter import *
try:
  import Pmw
except:
  import PmwLocal as Pmw
import tkFileDialog, tkSimpleDialog, tkMessageBox
import string, operator, traceback, sys, getopt, os, time
import vs, vsprogram, textedit, db
from types import *
import copy, pickle
import math
import mutex
from interaction import *
import entropy
import LogWin
import yaml # PERCY

SMARTEDIT_VERSION = '1.0'
SCREEN_SHOT_MODE = 0
DEMO_MODE = 1
DEBUG_MODIFIERS = 0
DEBUG = 0
INNER_LOOP_CONTROL_BAR = 0

USE_DIAMAND = 0

RUN_CONTROL_PANEL = 0

# Whether to enable startless programs
# PROGRAM_TYPE stores the default.
PROGRAM_TYPES = {'Fixed length program' : vsprogram.VSProgram,
        'Auto-segmented program' : vsprogram.VSSegmentlessProgram,
        'Auto-started/segmented program' : vsprogram.VSStartlessProgram,
        }
PROGRAM_TYPE = 'Fixed length program'

MODIFIERS = ['Control_L', 'Control_R', 'Shift_L', 'Shift_R', 'Alt_L', 'Alt_R']
MOD_MAPPING = {
  'Control_L':'Control',
  'Control_R':'Control',
  'Shift_L':'Shift',
  'Shift_R':'Shift',
  'Alt_L':'Mod1',
  'Alt_R':'Mod1'}
PUNCT = [
  "slash",
  "question",
  "equal",
  "plus",
  "minus",
  "underscore",
  "backslash",
  "bar",
  "bracketleft", "bracketright",
  "braceleft", "braceright",
  "semicolon", 'colon',
  'asciitilde',
  'quotedbl',
  'comma', 'less',
  'period', 'greater',
  'exclam', 'at', 'numbersign', 'dollar', 'percent',
  'asciicircum', 'ampersand', 'asterisk', 'parenleft', 'parenright',
  'Tab', 'space', 'Return'
  ]
if sys.platform != 'win32':
  PUNCT = PUNCT + [
  'grave',
  'apostrophe', 
  ]

MOVEMENT_COMMANDS = ['a', 'b', 'e', 'f', 'n', 'p']

DELETION_COMMANDS = ['d', 'h', 'k', 'w', 'x']

INSERTION_COMMANDS = ['y', 'v']

MISC_COMMANDS = ['t']

SELECTION_COMMANDS = ['slash']

COPY_COMMANDS = ['c']

# An exception to indicate that the program ended during "execution" of
# a program statement (that is, the statement suggested terminating the
# program).
try:
    class EndProgram(Exception):
     pass
except TypeError:
    # string based exceptions
    Empty = 'collect.EndProgram'

def mark_split(mark):
  return map(int, string.split(mark, '.'))

def mark_subtract(p1, p2):
  line1, char1 = mark_split(p1)
  line2, char2 = mark_split(p2)
  return '%i.%i' % (line1-line2, char1-char2)

def mark_difference(widget, p1, p2):
  if widget.compare(p1, '<', p2):
    return mark_subtract(p2, p1)
  else:
    return mark_subtract(p1, p2)

def mark_add(p1, p2):
  line1, char1 = mark_split(p1)
  line2, char2 = mark_split(p2)
  return '%i.%i' % (line1+line2, char1+char2)

def mark_greater(p1, p2):
  line1, char1 = mark_split(p1)
  line2, char2 = mark_split(p2)
  return (line1 > line2) or ((line1 == line2) and (char1 > char2))

def mark_create((row, col)):
  return '%i.%i' % (row, col)

def drag_select(text, start, delta):
  start = mark_add(text.index('insert'), start)
  end = mark_add(start, delta)
  text.tag_remove('sel', '1.0', 'end')
  text.tag_add('sel', start, end)
  text.mark_set('insert', end)
  return

def tkTextClosestGap(widget, x, y):
  pos = widget.index('@%i,%i' % (x, y))
  bbox = widget.bbox(pos)
  if not bbox:
    return pos
  if x - bbox[0] < bbox[2]/2:
    return pos
  return widget.index('%s + 1 char' % pos)

class ShowHypothesis(Toplevel):
  def __init__(self, title, func):
    Toplevel.__init__(self)
    self.title(title)

    b = Button(self, text='Close', command=self.destroy)
    b.pack(side=BOTTOM, expand=0, fill=X)
    b = Button(self, text='Update', command=self.update_hyp)
    b.pack(side=BOTTOM, expand=0, fill=X)

    self.text = Pmw.ScrolledText(self)
    self.text.pack(side=TOP, expand=1, fill=BOTH)
    hyp = func()
    self.func = func
    self.text.insert('end', hyp)
  
  def update_hyp(self):
    self.text.delete('1.0', 'end')
    self.text.insert('end', self.func())

class ProgramState:
  def __init__(self, cursorpos, text):
    self.set_defaults()

    # Take one character less than the whole file to cut off
    # the spurious end-line.
    #self.contents = text.get('1.0', 'end - 1c')
    self.contents = text.get('1.0', 'end - 1c').encode('utf-8') # PERCY
    #print "ProgramState", len(self.contents)
    self.pos = len(text.get('1.0', cursorpos))
    try:
      #self.clipboard = text.selection_get(selection='CLIPBOARD')
      self.clipboard = text.selection_get(selection='CLIPBOARD').encode('utf-8') # PERCY
    except TclError:
      self.clipboard = ''
    # This will be either an empty tuple (if there is no selection
    # defined), or a two-tuple of text widget position strings
    #self.selection_extent = text.tag_ranges('sel')
    # PERCY: tag_ranges now returns textindex objects; this program assumes strings
    se = text.tag_ranges('sel')
    if len(se) > 0:
      se = (str(se[0]), str(se[1]))
    #print "SEL", se
    self.selection_extent = se
  
  def set_defaults(self):
    self.contents = ''
    self.pos = '1.0'
    self.clipboard = ''
    self.selection_extent = ()
    # The iteration number is set inside VSProgram.
    self.iteration_number = -1

  def __setstate__(self, dict):
    self.set_defaults()
    for key in dict.keys():
      self.__dict__[key] = dict[key]
#    self.__dict__ = copy.copy(dict)
#    if not self.__dict__.has_key('clipboard'):
#      self.clipboard = ''
#    if not self.__dict__.has_key('selection_extent'):
#      self.selection_extent = ()
  def __repr__(self):
    return 'Pos: %s\nClipboard: <%s>\nSelection: <%s>\nIteration: %i\nEND' %\
      (`self.pos`, `self.clipboard`, `self.selection_extent`,
       self.iteration_number)
  def __cmp__(self, other):
    # Shortcut identical objects
    if self is other:
      return 0
    
    cmp_contents = cmp(self.contents, other.contents)
    if cmp_contents != 0:
      return cmp_contents
    else:
      return cmp(self.pos, other.pos)

class Event:
  """Class encapsulating information about a low-level event.  To be
  used for recording/replaying events in order to test our learning
  algorithms in batch mode."""
  def __init__(self, seq, timedelta):
    # Seq should be an event sequence such as "<KeyPress-a>"
    self.sequence = seq
    # Timedelta should be the time since the last event
    self.timedelta = timedelta
    self.dict = {}
  
  def __str__(self):
    return self.sequence + ' ' + `self.dict`
  
  def __setitem__(self, var, val):
    self.dict[var] = val
  
  def __getitem__(self, var):
    return self.dict[var]

  def generate(self, widget):
    newdict = copy.copy(self.dict)
    if newdict.has_key('pos'):
      pos = newdict['pos']
      del newdict['pos']
      widget.see(pos)
      (x0, y0, w, h) = widget.bbox(pos)
      x = x0
      y = y0
      newdict['x'] = x
      newdict['y'] = y

    #print "generate", self.sequence, newdict
    # PERCY: work around stupid bug where <Shift-Key-less> rewrites as '>'
    if self.sequence == '<Shift-Key-less>': self.sequence = '<Shift-Key-comma>'
    #if self.sequence == '<Key-Tab>': self.sequence = '<KeyPress-Tab>'
    # HACK: for some reason, tab doesn't trigger insert_string, so have to do it manually;
    # STILL NOT GOOD ENOUGH because first tab does not register; wtf???
    if self.sequence == '<Key-Tab>':
      class A: pass
      e = A()
      e.keysym = self.sequence
      app.insert_string(e)
    apply(widget.event_generate, (self.sequence,), newdict)

app = None

class MyText(Text):
  # Class to override certain methods of Text
  def __init__(self, master=None, cnf={}, **kw):
    apply(Text.__init__, (self, master, cnf), kw)
    self.mutex = mutex.mutex()
  
  def execute_and_release(self, event, func):
    func(event)
    self.mutex.unlock()

  def lock_call(self, event, func, retval):
    if func is None:
      func = lambda e: None
    
    self.mutex.lock(lambda e, s=self, f=func:
           s.execute_and_release(e, f), event)
    return retval

  def bind(self, sequence=None, func=None, add=None, retval=None):
    """Call Text.bind but insert a mutex wrapper around the event's
    callback.  Retval is not in the Text.bind interface; it denotes
    the return value for the callback (either None or 'break')."""
    # Add locking on to this function call.
    Text.bind(self, sequence,
        lambda e, s=self, f=func, r=retval: s.lock_call(e, f, r),
        add)

  def bind_class(self, className, sequence=None, func=None, add=None,
      retval=None):
    """See the documentation for bind()"""
    # Add locking on to this function call.
    Text.bind_class(self, className, sequence,
           lambda e, s=self, f=func, r=retval:
           s.lock_call(e, f, r),
           add)

class App(Frame):
  def __init__(self, master=None):
    Frame.__init__(self, master)
    self.master.title('SMARTedit')
    self.__events = []
    self.__abstract_events = []
    self.__history = []
    self._collect = 0    # whether to record events
    self.__modifiers = {}  # active modifier keys
    # default fonts to display
    if DEMO_MODE:
      self.default_font_size = 2
    else:
      self.default_font_size = 0
    self.button_down = 0
    self.pack(expand=1, fill=BOTH)
    self.bindings = []
    #------------------------------------------------------------
    # Program state
    self.lastaction = None
    self.trackingundostate = None

    #------------------------------------------------------------
    # Undo info
    self.max_undo_steps = 5
    self.undo_trace = []
    self.current_undo_step = None

    #------------------------------------------------------------
    self.filename = None
    self.current_action = None
    # Running scenarios
    self.scenario_proc = None
    # Our version space
    try:
      self.vs = PROGRAM_TYPES[PROGRAM_TYPE]()
    except KeyError:
      raise RuntimeError,\
        'Unknown value for PROGRAM_TYPE: %s' % `PROGRAM_TYPE`
    # Enable/disable online learning
    # I believe this cannot be disabled, otherwise things will break
    # 9/26/00 tlau
    self.online_learning = 1
    # Undo state for resetting to before tentative action
    self.undostate = None
    # Enable/disable active learning
    self.active_learning = 0

    # For playback/run
    self.playback_speed = DoubleVar()
    self.do_continue_interruptable_run = 0

    # Create our widgets
    self._create_widgets()
    # Must get text state after creating the text widget
    self.last_state = self.get_text_state('1.0')
    self.first_state = self.last_state
    # Create log window
    self.logwin = LogWin.LogWin('SMARTedit messages:', self)
    self.logwin.withdraw()

    self.choice_popup = None
  
    self.user = None

    # Info window
#    self.controlpanel = None

    # Keypress/mouseclick counters
    self.keypresses = 0
    self.mouseclicks = 0
    if DEBUG:
      sys.exitfunc = self.show_counters_text

    self.curr_scenario = db.Scenario(self.filename)
    self.high_level_actions = [] # PERCY: start recording (None marks examples)

  def _create_widgets(self):
    menubar = Frame(self, relief='raised', borderwidth=2)
    if not SCREEN_SHOT_MODE:
      menubar.pack(side=TOP, fill=X, expand=0)
    filemenu = Menubutton(menubar, text='File', underline=0)
    filemenu.pack(side='left')
    filemenu.menu = Menu(filemenu, tearoff=0)
    filemenu.menu.add_command(label='Reset',
      command=self.reset, underline=0)
    filemenu.menu.add_command(label='Open', command=self.fileopen,
      underline=0)
    filemenu.menu.add_command(label='Save', command=self.filesave,
      underline=0)
    filemenu.menu.add_command(label='Dump scenario', command=self.dumpScenario,
      underline=0)
    filemenu.menu.add_command(label='Save as',
      command=self.filesaveas, underline=5)
    filemenu.menu.add_command(label='Save profile',
      command=self.save_profile, underline=5)
    filemenu.menu.add_command(label='Choose profile...',
      command=self.choose_profile, underline=0)
    filemenu.menu.add_separator()

    # Font size menu
    self.fontsize_var = IntVar()
    self.fontsize_var.set(self.default_font_size)
    fontsize = Menu(filemenu.menu, tearoff=0)
    fontsize.add_radiobutton(label='Small',
      command=self.change_fontsize, underline=0,
      variable=self.fontsize_var, value=0)
    fontsize.add_radiobutton(label='Medium',
      command=self.change_fontsize, underline=0,
      variable=self.fontsize_var, value=1)
    fontsize.add_radiobutton(label='Large',
      command=self.change_fontsize, underline=0,
      variable=self.fontsize_var, value=2)
    fontsize.add_radiobutton(label='Huge',
      command=self.change_fontsize, underline=0,
      variable=self.fontsize_var, value=3)
    fontsize.add_radiobutton(label='Enormous',
      command=self.change_fontsize, underline=0,
      variable=self.fontsize_var, value=4)

    filemenu.menu.add_cascade(label='Font size...',
      command=self.filesaveas, underline=0, menu=fontsize)
    filemenu.menu.add_separator()
    filemenu.menu.add_command(label='Exit', command=self.quit,
      underline=1)
    filemenu.config(menu=filemenu.menu)

    editmenu = Menubutton(menubar, text='Edit', underline=0)
    editmenu.pack(side='left')
    editmenu.menu = Menu(editmenu, tearoff=0)
    editmenu.menu.add_command(label='Cut (C-x)', command=self.editcut,
      underline=0)
    editmenu.menu.add_command(label='Copy (C-c)', command=self.editcopy,
      underline=0)
    editmenu.menu.add_command(label='Paste (C-y)', command=self.editpaste,
      underline=0)
    editmenu.menu.add_command(label='Search (C-s)',
      command=self.editsearch, underline=0)
    editmenu.config(menu=editmenu.menu)

    programmenu = Menubutton(menubar, text='Program', underline=0)
    programmenu.pack(side='left')
    programmenu.menu = Menu(programmenu, tearoff=0)
    programmenu.menu.add_command(label='Clear program',
                  command=self.abort,
                  underline=0)
    programmenu.menu.add_command(label='Backup program',
                  command=self.save_program,
                  underline=0)
    programmenu.menu.add_command(label='Restore program',
                  command=self.load_program,
                  underline=0)
    programmenu.menu.add_command(label='Save program as...',
                  command=self.save_program_as,
                  underline=0)
    programmenu.menu.add_command(label='Load program from...',
                  command=self.load_program_from,
                  underline=0)
    programmenu.config(menu=programmenu.menu)

    optionmenu = Menubutton(menubar, text='Options', underline=0)
    optionmenu.pack(side='left')
    optionmenu.menu = Menu(optionmenu, tearoff=0)
    # Program type
    self.var_progtype = StringVar()
    self.var_progtype.set(PROGRAM_TYPE)
    prog_types = PROGRAM_TYPES.keys()
    prog_types.sort()
    for str in prog_types:
      optionmenu.menu.add_radiobutton(label=str,
                    command=self.set_progtype,
                    var=self.var_progtype,
                    value=str)

    optionmenu.menu.add_separator()
    # Active learning
    self.var_active = BooleanVar()
    self.var_active.set(self.active_learning)
    optionmenu.menu.add_checkbutton(label='Active learning',
      command=self.set_active, var=self.var_active)
    # NO_INAPP_ADJUSTMENT
    self.var_inappadj = BooleanVar()
    self.var_inappadj.set(not vs.NO_INAPP_ADJUSTMENT)
    optionmenu.menu.add_checkbutton(
      label='Lower inapplicable probabilities',
      command=self.set_inapp_adj, var=self.var_inappadj)
    # NO_INAPP_ALLOWED
    self.var_inappallow = BooleanVar()
    self.var_inappallow.set(vs.NO_INAPP_ALLOWED)
    optionmenu.menu.add_checkbutton(
      label='Ignore inapplicable hypotheses',
      command=self.set_inapp_allow, var=self.var_inappallow)

    optionmenu.menu.add_separator()
    # ALWAYS_NORMALIZE
    self.var_normalize = BooleanVar()
    self.var_normalize.set(vs.ALWAYS_NORMALIZE)
    optionmenu.menu.add_checkbutton(
      label='Always normalize',
      command=self.set_normalize, var=self.var_normalize)
    # USE_DIAMAND
    self.var_diamand = BooleanVar()
    self.var_diamand.set(USE_DIAMAND)
    optionmenu.menu.add_checkbutton(
      label='Use Diamand',
      command=self.set_diamand, var=self.var_diamand)
    # inner loop bar
    self.var_loopbar = BooleanVar()
    self.var_loopbar.set(INNER_LOOP_CONTROL_BAR)
    optionmenu.menu.add_checkbutton(
      label='Show inner loop buttons',
      command=self.set_loopbar, var=self.var_loopbar)

    # Enable the option menu
    optionmenu.config(menu=optionmenu.menu)

    if DEBUG:
      debugmenu = Menubutton(menubar, text='Debug', underline=0)
      debugmenu.pack(side='left')
      debugmenu.menu = Menu(debugmenu, tearoff=0)
      debugmenu.menu.add_command(label='Show event counters',
        command=self.show_counters_popup, underline=0)
#    debugmenu.menu.add_command(label='Reset event counters',
#      command=self.reset_counters, underline=0)
      debugmenu.config(menu=debugmenu.menu)

    scenariomenu = Menubutton(menubar, text='Scenarios', underline=0)
    scenariomenu.pack(side='left')
    scenariomenu.menu = Menu(scenariomenu, tearoff=0)
    #scenariomenu.menu.add_command(label='Hijacked! Test choice.',
    #  command=self.entropy_choice, underline=0)
    scenariomenu.menu.add_command(label='Open scenario',
      command=self.open_scenario, underline=0)
    scenariomenu.menu.add_command(label='Test all scenarios',
      command=self.test_all_scenarios, underline=0)
    scenariomenu.menu.add_command(label='Dump all scenarios',
      command=self.dump_all_scenarios, underline=0)
    scenariomenu.config(menu=scenariomenu.menu)

    # Help menu
    helpmenu = Menubutton(menubar, text='Help', underline=0)
    helpmenu.pack(side='right')
    helpmenu.menu = Menu(helpmenu, tearoff=0)
    helpmenu.menu.add_command(label='Status bar history',
      command=self.show_status_history, underline=0)
    helpmenu.menu.add_command(label='About', command=self.about)
    helpmenu.config(menu=helpmenu.menu)

    buttonbar = Frame(self, relief='groove', borderwidth=2)
    if not USE_DIAMAND:
      buttonbar.pack(side=TOP, expand=0, fill=X)
    if SCREEN_SHOT_MODE:
      self.macrobut = Button(buttonbar, text='Record\niteration',
        command=self.start_collect)
      self.undobut = Button(buttonbar, text='Undo',
        command=self.undo)
      self.stepbut = Button(buttonbar, text='Step through\nmacro',
        command=self.step_vector)
      self.guessbut = Button(buttonbar, text='Try another\nguess',
        command=self.guess_again, state='disabled')
      self.jumpbut = Button(buttonbar, text='Jump\nexample',
        command=self.jump)
      self.donebut = Button(buttonbar, text='No more\nexamples',
        command=self.no_more_examples)
      self.intrunbut = Button(buttonbar, text='Run\nprogram',
                  command=lambda s=self:
                  s.interruptable_run((1.5,0.2,0.1)))
    else:
      self.macrobut = Button(buttonbar, text='Record iteration',
        command=self.start_collect)
      self.undobut = Button(buttonbar, text='Undo',
        command=self.undo)
      self.stepbut = Button(buttonbar, text='Step through macro',
                command=self.step_vector)
      self.guessbut = Button(buttonbar, text='Try another guess',
        command=self.guess_again, state='disabled')
      self.jumpbut = Button(buttonbar, text='Jump to example',
        command=self.jump)
      self.donebut = Button(buttonbar, text='No more examples',
        command=self.no_more_examples)
      self.intrunbut = Button(buttonbar, text='Run program',
                  command=lambda s=self:
                  s.interruptable_run((1.5, 0.2,
                  0.1)))

    self.macro_buttons = [self.macrobut,
              # self.undobut, # UNDO IS NOT WORKING
              self.stepbut,
              self.guessbut,
#              self.donebut,
              self.intrunbut]
    for but in self.macro_buttons:
      but.pack(side=LEFT, expand=0, fill=X, pady=3, padx=3)
    self.buttonbar = buttonbar

    self.loopbar = Frame(self, relief='groove', borderwidth=2)
    self.inner_loop_start_but = Button(self.loopbar,
      text='Start inner loop', command=self.inner_start)
    self.inner_loop_start_but.pack(side=LEFT, expand=0, fill=X,
      pady=3, padx=3)
    self.inner_loop_stop_but = Button(self.loopbar,
      text='Finished inner loop', command=self.inner_done,
      state='disabled')
    self.inner_loop_stop_but.pack(side=LEFT, expand=0, fill=X,
      pady=3, padx=3)
    self.macro_buttons = self.macro_buttons +\
      [self.inner_loop_start_but, self.inner_loop_stop_but]
    if INNER_LOOP_CONTROL_BAR and not USE_DIAMAND:
      self.loopbar.pack(after=self.buttonbar, expand=0, fill=X)

#    self.dobut = Button(buttonbar, text='Run macro',
#      command=self.replay, state='disabled')
#    self.dobut.pack(side=LEFT, expand=0, fill=X, padx=3, pady=3)

    # Modifier labels
    modbar = Frame(self, relief='groove', borderwidth=2)
    if DEBUG_MODIFIERS:
      modbar.pack(side=TOP, expand=1, fill=X)
    for key in MODIFIERS:
      self.__modifiers[key] = 0
      l1 = Label(modbar, text=key + ':')
      l1.pack(side='left')
      exec('self.%s_var = BooleanVar()' % key)
      var = eval('self.%s_var' % key)
      var.set(0)
      l2 = Label(modbar, textvariable=var)
      l2.pack(side='left')

    # Status bar
    self._status = Label(self, anchor='w', relief='groove')
    self._status.pack(side=TOP, fill=X, expand=0)
    self.status_cb = None

    # Interaction panel
    interactions = [
      RecordIterationInteraction,
      RecordActionInteraction,
      SuggestActionInteraction,
      PlaybackInteraction,
      RunWhileConfidentInteraction,
      JumpInteraction,
      ]
    controllers = [
      RecordController,
      RecordController,
      SuggestController,
      PlaybackController,
      RunWhileConfidentController,
      JumpController,
      ]
    names = [
      'Record one full example',
      'Record one step',
      'Step',
      'Run to end of example',
      'Run while sure',
      'Jump and record an example',
      ]

    self.interaction_panel = IIBarsAndFading(self)
    for i in range(len(interactions)):
      interaction = interactions[i]()
      controller = controllers[i](self, interaction)
      self.interaction_panel.add_interaction(interaction,
                        controller,
                        names[i])

    if USE_DIAMAND:
      self.interaction_panel.pack(side=LEFT, expand=0, fill=Y)
      self.interaction_panel.initialize()
      self.choose_profile('tlau')

    # Text widget
    if SCREEN_SHOT_MODE:
      t = MyText(self,width=40, height=5, relief='flat',
           highlightthickness=0, wrap='word', exportselection=1,
           insertwidth=15,
           insertbackground='blue', insertofftime=0)
    else:
      t = MyText(self,width=80, height=25, relief='flat',
           highlightthickness=0, wrap='word', exportselection=1,
           insertwidth=5,
           insertbackground='blue', insertofftime=100)
    t.tag_configure('sel', background='cyan')
    sb = Scrollbar(self, orient='vertical', command=t.yview)
    t.configure(yscrollcommand=sb.set)
    sb.pack(side=RIGHT, expand=0, fill=Y)

    # PERCY
    sb = Scrollbar(self, orient='horizontal', command=t.xview)
    t.configure(xscrollcommand=sb.set)
    sb.pack(side=BOTTOM, expand=0, fill=X)
    t.configure(wrap='none') # PERCY

    t.pack(side=TOP, expand=1, fill=BOTH)
    self.text = t

    # Remap C-v to paste
    self.text.bind('<Control-Key-v>', self.paste_break,
      retval='break')
    # Set up keybindings
    self.bindkeys_macro()

    # Set up event counters
    self.master.bind('<KeyPress>', self.count_key_press, add=1)
    self.master.bind('<ButtonPress>', self.count_button_press, add=1)

    # Set up default fonts
    self.fontsize_var.set(self.default_font_size)
    self.change_fontsize()
  
  # Pop up an about box
  def about(self):
    Pmw.aboutversion(SMARTEDIT_VERSION)
    Pmw.aboutcopyright('Copyright Tessa Lau 200j')
    Pmw.aboutcontact(
      'Tessa Lau\n' +
      'tlau-smartedit@ofb.net\n' +
      'http://ofb.net/~tlau/research/smartedit/')
    self.about = Pmw.AboutDialog(self, applicationname='SMARTedit')
    self.about.show()

  # Pop up the history of status messages
  def show_status_history(self):
    if hasattr(self, 'logwin'):
      self.logwin.deiconify()
      self.logwin.tkraise()

  def paste_break(self, event):
    self.text.event_generate('<<Paste>>')
    return 'break'

  def loadfile(self, fname):
    self.filename = fname
    try:
      input = open(fname, 'r')
      content = input.read()
      input.close()
    except:
      traceback.print_exc(1)
    self.loadcontent(content)

  def loadcontent(self, content):
    self.text.delete('1.0', 'end')
    self.text.insert('end', content)
    self.text.mark_set('insert', '1.0')
    self.text.focus_set()
    if self.filename:
      self.master.title('SMARTedit: %s' %
        os.path.split(self.filename)[1])
  
  def gotoline(self, lineno):
    self.text.mark_set('insert', '%s.0' % lineno)
    self.text.see('insert')

  def gotocol(self, colno):
    (row, col) = mark_split(self.text.index('insert lineend'))
    if col > colno:
      colno = col
    self.text.mark_set('insert', 'insert linestart + %s chars' % colno)
    self.text.see('insert')

  def reset(self):
    if self.filename and os.path.exists(self.filename):
      self.loadfile(self.filename)
    else:
      self.text.delete('1.0', 'end')
      self.filename = None
    self.clear_events()
  
  # ----------------------------------------------------------------------
  # Option variables
  def set_inapp_adj(self):
    vs.NO_INAPP_ADJUSTMENT = not self.var_inappadj.get()
  def set_inapp_allow(self):
    vs.NO_INAPP_ALLOWED = self.var_inappallow.get()
  def set_normalize(self):
    vs.ALWAYS_NORMALIZE = self.var_normalize.get()
  def set_diamand(self):
    global USE_DIAMAND
    USE_DIAMAND = self.var_diamand.get()
    if USE_DIAMAND:
      self.buttonbar.forget()
      if INNER_LOOP_CONTROL_BAR:
        self.loopbar.forget()
      self.interaction_panel.pack(side=LEFT, expand=0, fill=Y,
        before=self.text)
      self.interaction_panel.initialize()
      self.choose_profile('tlau')
    else:
      self.interaction_panel.forget()
      self.buttonbar.pack(before=self.text, side=TOP, expand=0,
        fill=X)
      if INNER_LOOP_CONTROL_BAR:
        self.loopbar.pack(after=self.buttonbar, expand=0,
          fill=X)
    self.clear_events()
  def set_loopbar(self):
    global INNER_LOOP_CONTROL_BAR
    INNER_LOOP_CONTROL_BAR = self.var_loopbar.get()
    if INNER_LOOP_CONTROL_BAR and not USE_DIAMAND:
      self.loopbar.pack(after=self.buttonbar, expand=0, fill=X)
    else:
      self.loopbar.forget()
  def set_progtype(self):
    global PROGRAM_TYPE
    PROGRAM_TYPE = self.var_progtype.get()
    # For speed purposes, auto-segmented programs can grow to
    # any length (-1 --> infinity), but auto-started programs
    # get set to a default maximum length.
    if PROGRAM_TYPE == 'Auto-segmented program':
      vs.MAX_PROGRAM_LENGTH = -1
    elif PROGRAM_TYPE == 'Auto-started/segmented program':
      vs.MAX_PROGRAM_LENGTH = vs.MAX_PROGRAM_LENGTH_DEFAULT

    if PROGRAM_TYPE not in PROGRAM_TYPES.keys():
      raise RuntimeError,\
        'Unsupported value for program type: %s' % `PROGRAM_TYPE`
    # Need to clear state here, to reset to correct type of program
    self.clear_events()
    
  def set_active(self):
    self.active_learning = self.var_active.get()
  
  # ----------------------------------------------------------------------
  # Font size changing
  def change_fontsize(self):
    sz = self.fontsize_var.get()
    if sz == 0:
      self.text.configure(font='Courier 10 bold')
      self._status.configure(font='Courier 10 bold')
      for b in self.macro_buttons:
        b.configure(font='Helvetica 10')
    elif sz == 1:
      self.text.configure(font='Courier 12 bold')
      self._status.configure(font='Courier 12 bold')
      for b in self.macro_buttons:
        b.configure(font='Helvetica 12')
    elif sz == 2:
      self.text.configure(font='Courier 14 bold', width=70)
      self._status.configure(font='Courier 14 bold')
      for b in self.macro_buttons:
        b.configure(font='Helvetica 14')
    elif sz == 3:
      self.text.configure(font='Courier 18 bold', width=70)
      self._status.configure(font='Courier 18 bold')
      for b in self.macro_buttons:
        b.configure(font='Helvetica 18')
    elif sz == 4:
      self.text.configure(font='Courier 24 bold', width=70)
      self._status.configure(font='Courier 24 bold')
      for b in self.macro_buttons:
        b.configure(font='Helvetica 24')
    else:
      print 'Unknown font size selected', sz

  # ------------------------------------------------------------
  # Take a snapshot of the text state around this cursor position
  def get_text_state(self, pos):
    state = ProgramState(pos, self.text)
    return state
  
  # Set the text to this state
  def set_text_state(self, state):
    self.text.delete('1.0', 'end')
    self.text.insert('end', state.contents)
    self.text.mark_set('insert',
              mark_create(textedit.rowColTuple(state.contents,
                          state.pos)))
    self.text.tag_remove('sel', '1.0', 'end')
    if state.selection_extent != ():
      self.text.tag_add('sel',
               state.selection_extent[0],
               state.selection_extent[1])
    self.text.clipboard_clear()
    self.text.clipboard_append(state.clipboard)
    self.text.see('insert')
  
  # ------------------------------------------------------------
  # Generic text editor menu functions
  def fileopen(self, filename=None):
    if filename is None:
      filename = tkFileDialog.askopenfilename()
    if filename is not None and filename:
      self.loadfile(filename)
    self.clear_events()

  # For creating new scenarios.  Record some action and then just save it.
  def dumpScenario(self): # PERCY
    if self.filename is not None:
      import re
      pathPrefix = re.sub("\\.\w+$", "", self.filename)
      self.dumpCurrScenarioToDisk(pathPrefix)

  def dumpCurrScenarioToDisk(self, pathPrefix): # PERCY
    map = {}
    map['name'] = os.path.basename(pathPrefix)
    map['examples'] = []

    # Fetch the examples that we've recorded
    exmap = None
    #print self.high_level_actions
    for triple in self.high_level_actions:
      if triple is None: # End of this example
        print " ", repr(" | ".join(exmap['actions'])) # New example
        map['examples'].append(exmap)
        exmap = None
      else:
        action, instate, outstate = triple
        if exmap is None:
          exmap = {}
          exmap['before'] = self.convertState(instate)
          exmap['actions'] = []
        exmap['after'] = self.convertState(outstate) # Will get overridden by the last one
        exmap['actions'].append(self.convertAction(action, instate, outstate))
    assert (exmap is None) # Must end with None to flush
    self.writeScenarioYamlExamples(pathPrefix, map)

  def filesave(self):
    if not self.filename:
      self.set_status('Current file is unnamed.  Use Save As ' +\
        'instead.')
      return

    try:
      f = open(self.filename, 'w')
      f.write(self.text.get('1.0', 'end - 1 ch'))
      f.close()
    except IOError, e:
      self.set_status('Error writing file %s: %s' %
        (self.filename, str(e)))

  def filesaveas(self):
    filename = tkFileDialog.asksaveasfilename()
    if filename is not None:
      self.filename = filename
      self.filesave()

  def save_program_as(self):
    types = [('programs', '*.prg'),
          ('all files', '*')]
    filename = tkFileDialog.asksaveasfilename(filetypes=types)
    if filename is not None:
      self.write_program(filename, self.vs)

  def load_program_from(self, filename=None):
    if filename is None:
      types = [('programs', '*.prg'),
            ('all files', '*')]
      filename = tkFileDialog.askopenfilename(filetypes=types)
    if filename is not None and filename:
      self.vs = self.read_program(filename)
      self.vs.action_number = 0

  def editcut(self):
    try:
      sel = self.text.get('sel.first', 'sel.last')
    except TclError:
      return
    self.text.event_generate('<<Cut>>')

  def editcopy(self):
    try:
      sel = self.text.get('sel.first', 'sel.last')
    except TclError:
      return
    self.text.event_generate('<<Copy>>')

  def editpaste(self):
    self.text.event_generate('<<Paste>>')
  
  def editsearch(self):
    import tkSimpleDialog
    searchstr = tkSimpleDialog.askstring('SMARTedit search',
                    'Search for string:')
    self.text.focus_set()
    if not searchstr:
      return
    loc = self.text.search(searchstr, 'insert')
    if loc == '':
      self.set_status("Couldn't find string %s." % `searchstr`)
    else:
      self.text.mark_set('insert', loc)
      self.text.see(loc)
      self.add_action('move', 'move')
    
  # ------------------------------------------------------------
  
  # ------------------------------------------------------------
  # Status bar
  def _munge_for_display(self, msg):
    if not msg:
      return '<EMPTY STRING>'

    replace_patterns = [
      ('\\012', '\\n'),
      ('\\\\',  '\\'),
      ]

    new_msg = msg
    for pair in replace_patterns:
      new_msg = string.replace(new_msg, pair[0], pair[1])
      
    return new_msg
          
  def set_status(self, msg, guaranteed_time = 0):
    if hasattr(self, 'logwin'):
      self.logwin.add_message(msg)
    if hasattr(self, 'waiting') and self.waiting > 0:
      self.after(int(self.waiting * 1000),
           lambda s=self,m=msg,g=guaranteed_time:
           s.set_status(m,g))
      return
    self._status['background'] = 'green'
    msg = self._munge_for_display(msg)
    if DEMO_MODE:
      maxmsglen = 69
    else:
      maxmsglen = 80
    if len(msg) >= maxmsglen:
      msg = msg[:maxmsglen-1] + '...'
    self._status['text'] = msg
    if guaranteed_time:
      self.waiting = guaranteed_time
      def zero_wait(s=self):
        s.waiting = 0
      self.after(int(guaranteed_time * 1000), zero_wait)
    if SCREEN_SHOT_MODE or DEMO_MODE: return
    if self.status_cb is not None:
      self.after_cancel(self.status_cb)
    self.status_cb = self.after(2000, self.normal_status)

  def normal_status(self):
    oldbg = self._status.configure('background')[3]
    self._status['background'] = oldbg
    self.status_cb = None
  # ------------------------------------------------------------
  
  # ------------------------------------------------------------
  # Manipulate high-level actions
  def create_action(self, action, instate, outstate):
    # Prepare for an undo.
    # self.save_undo_step() # UNDO IS NOT WORKING
    if DEBUG: print "create_action"

    if action == 'move':
      e = textedit.Move(instate, outstate)
    elif action == 'select':
      e = textedit.Select(instate, outstate)
      #print "SEL", repr(outstate.selection_extent)
    elif action == 'paste':
      #print "PASTE", repr(outstate.clipboard), len(instate.contents), len(outstate.contents)
      e = textedit.Paste(instate, outstate)
    elif action == 'cut':
      e = textedit.Cut(instate, outstate)
      #print "CUT", repr(instate.selection_extent), repr(outstate.clipboard)
    elif action == 'copy':
      e = textedit.Copy(instate, outstate)
    elif action == 'delete':
      e = textedit.Delete(instate, outstate)
    elif action == 'insert':
      #print "INSERT", len(instate.contents), len(outstate.contents)
      e = textedit.Insert(instate, outstate)
    elif action == 'deleteselection':
      e = textedit.DeleteSelection(instate, outstate)
      #print "DEL", repr(outstate.clipboard)
    else:
      print 'Unknown action', action
      return
    try:
      self.vs.feed_nth(e)
    except vs.VSError, e:
      self.set_status('Error: ' + str(e))

    #print "HIGH", self.high_level_actions
    self.addHighLevelAction((action, instate, outstate))
    #self.curr_scenario.add_iteration(instate, events, afterstate) # XXX

  def addHighLevelAction(self, triple):
    #if hasattr(self, 'high_level_actions'): # PERCY
    if triple is None: print "ADD_TO_HIGH: None"
    else: print "ADD_TO_HIGH:", triple[0]
    self.high_level_actions.append(triple)

  # Given a position in a buffer, find what position it would correspond
  # to in the buffer obtained by undoing everything in change_list.
  def back_out_position(self, pos, change_list, direction = 1):
    change_list.reverse()

    for (change_type, loc, size) in change_list:
      if change_type == 'insert':
        if pos <= loc:
          # No effect, the change was later in the buffer
          pass
        elif loc < pos < loc + size:
          # The position is in the middle of the inserted
          # text, so the equivalent before insertion is at
          # the start
          pos = loc
        else:
          # All the inserted text is before the change
          pos = pos - size
      elif change_type == 'delete':
        if pos <= loc:
          # No effect, the change was later in the buffer
          pass
        elif loc == pos:
          # The position is on the deletion point, so it
          # was somewhere in the deleted text. We allow it
          # to be either before or after:
          if direction > 0:
            pos = loc + size
          else:
            pass
        else:
          # All the deleted text is before the change
          pos = pos + size
      else:
        raise "Unknown change type encountered."

    change_list.reverse()

    return pos

  # Keeps track of changes to the buffer for backing out locations in
  # the original buffer.
  def add_change(self, action, instate, outstate,
        change_list = None):
    if change_list is None:
      change_list = self.change_list
    NO_CHANGE_ACTIONS = ['move', 'select', 'copy']
    INSERT_ACTIONS = ['paste', 'insert']
    DELETE_ACTIONS = ['delete']
    DELETE_SELECTION_ACTIONS = ['deleteselection', 'cut']
    
    if self.change_list == None:
      self.change_list = []
      
    if action in NO_CHANGE_ACTIONS:
      pass
    elif action in INSERT_ACTIONS:
      self.change_list.append(('insert', instate.pos,
                len(outstate.contents) -
                len(instate.contents)))
    elif action in DELETE_ACTIONS:
      self.change_list.append(('delete', outstate.pos,
                len(instate.contents) -
                len(outstate.contents)))
    elif action in DELETE_SELECTION_ACTIONS:
      assert(instate.selection_extent != ())
      start_tuple = mark_split(instate.selection_extent[0])
      pos = textedit.positionFromTuple(instate.contents, start_tuple)
      self.change_list.append(('delete', pos,
                len(instate.contents) -
                len(outstate.contents)))
    else:
      raise 'Unsupported change type!'

  def add_action(self, lastaction, newaction):
    # Implement the state machine that translates from low-level
    # interface actions to high-level actions

    #DEBUG!!
    if DEBUG: print 'Adding action:', lastaction, newaction

    # If the new action is different from the old one, group
    # all recent events into one SMARTedit action.
    if lastaction is None or self.lastaction != lastaction:
      ## The aggregate action is changing. This is where to
      ## put anything to handle transitioning to a new action
      self.lastundostate = self.trackingundostate

      # First, create the old action.
      if self.lastaction is not None:
        # Dump old last action
        self.create_action(self.lastaction, self.first_state,
                  self.last_state)

      self.first_state = self.last_state

    self.last_state = self.get_text_state('insert')
    # self.record_undo_step() # UNDO IS NOT WORKING
    self.lastaction = newaction

    self.trackingundostate = self.getundostate()
    #DEBUG!!
    #print 'Done Adding action: New action:', self.lastaction

  # ------------------------------------------------------------

  def clear_events(self):
    self.disable_guess_button()
    self.__abstract_events = []
    try:
      self.vs = PROGRAM_TYPES[PROGRAM_TYPE]()
    except KeyError:
      raise RuntimeError,\
        'Unknown value for PROGRAM_TYPE: %s' % `PROGRAM_TYPE`
    self.lastaction = None
    self.first_state = self.get_text_state('1.0')
    self.last_state = self.first_state
    self.undostate = None
    self.playback_speed.set(4.0)
    if USE_DIAMAND:
      self.interaction_panel.reset()
    self.set_status('Ready')


  def do_nothing(self, event):
    self.set_status('Warning: key combination %s has no effect' %
      event.keysym)

  def add_binding_class(self, classname, event, cmd, **kw):
    """Create a class binding for classname from event to cmd"""
    b = apply(self.text.bind_class, (classname, event, cmd), kw)
    self.bindings.append((event, b))
    if DEBUG: print 'binding %s to' % event, cmd
  
  def add_binding(self, event, cmd, **kw):
    """Create a binding for event to cmd"""
    b = apply(self.text.bind, (event, cmd), kw)
    self.bindings.append((event, b))
    if DEBUG: print 'binding %s to' % event, cmd

  def bindkeys_macro(self):
    # remove all bindings
    for (seq, func) in self.bindings:
      self.text.unbind(seq, func)
    self.bindings = []

    for i in map(None, string.lowercase + string.uppercase +
      string.digits) + PUNCT:
      self.add_binding_class('MyPostClass', '<KeyPress-%s>' % i,
                self.insert_string)
      if i in MOVEMENT_COMMANDS:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.move_command, add=1)
      elif i in DELETION_COMMANDS:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.delete_command)
      elif i in INSERTION_COMMANDS:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.paste_command)
      elif i in MISC_COMMANDS:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.transpose_command)
      elif i in SELECTION_COMMANDS:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.extend_selection_command)
      elif i in COPY_COMMANDS:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.copy_selection)
      # Ctrl-digit is bound to one of the buttons.
      elif i in map(str, range(1,10)) and \
        int(i) <= len(self.macro_buttons):
        button = self.macro_buttons[int(i) - 1]
        
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  lambda e,but=button:
                  but.invoke())
      elif i == '0':
        self.add_binding_class('MyPostClass', '<Control-Key-0>',
                 lambda e, s=self:
                 s.advise_user())
      else:
        self.add_binding_class('MyPostClass',
                  '<Control-Key-%s>' % i,
                  self.do_nothing)
      if i == 'f':
        # Should really aquire a lock just like all other
        # functions that call add_action. !!!
        self.add_binding_class('MyPostClass',
                 '<Control-Shift-Key-f>',
                 lambda e, s=self:
                 s.add_action(None, None))
      else:
        self.add_binding_class('MyPostClass',
                  '<Control-Shift-Key-%s>' % i,
                  self.do_nothing)
      self.add_binding_class('MyPostClass',
                '<Mod1-Key-%s>' % i,
                self.do_nothing)
      self.add_binding_class('MyPostClass',
                '<Control-Mod1-Key-%s>' % i,
                self.do_nothing)
      self.add_binding_class('MyPostClass',
                '<Control-Shift-Mod1-Key-%s>' % i,
                self.do_nothing)
    self.add_binding_class('MyPostClass', '<Key-BackSpace>',
              self.delete_command)
    self.add_binding_class('MyPostClass', '<Key-Delete>',
              self.delete_command)
    self.add_binding_class('MyPostClass', '<Key-Left>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-Right>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-Up>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-Down>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-End>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-Home>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-Prior>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Key-Next>',
              self.move_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Left>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Up>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Down>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Right>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-End>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Home>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Prior>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Shift-Key-Next>',
              self.extend_selection_command)
    self.add_binding_class('MyPostClass', '<Control-Key-s>',
              lambda e, s=self:s.editsearch())

    self.add_binding_class('MyPostClass', '<<Paste>>',
              self.paste_command)

    self.add_binding_class('MyPostClass', '<<Cut>>',
              self.cut_command)

    self.add_binding_class('MyPostClass', '<<Copy>>',
              self.copy_selection)

    self.add_binding_class('MyPostClass', '<ButtonPress>',
              self.collect_button_press)
    self.add_binding_class('MyPostClass', '<KeyPress>',
              self.collect_key_press)
    self.add_binding_class('MyPostClass', '<KeyRelease>',
              self.collect_key_release)
    self.add_binding_class('MyPostClass', '<ButtonRelease-1>',
              self.collect_button_release)

    self.text.bindtags(self.text.bindtags() + ('MyPostClass',))

  def bindkeys_scenario(self):
    # remove all bindings
    for (seq, func) in self.bindings:
      try:
        self.text.unbind(seq, func)
      except TclError, e:
        print e
        print seq, func

    # remove the post-key-press bindings
    # for the moment, I'll assume that the last entry is the
    # one I want to remove. !!!
    assert(self.text.bindtags()[-1] == 'MyPostClass')
    self.text.bindtags(self.text.bindtags()[:-1])
    
    self.bindings = []

    self.add_binding('<ButtonPress>', self.scenario_button_press)
    self.add_binding('<Motion>', self.scenario_mouse_move)
    self.add_binding('<KeyPress>', self.scenario_key_press)
    self.add_binding('<KeyRelease>', self.scenario_key_release)
    self.add_binding('<ButtonRelease-1>', self.scenario_button_release)

  def selection_exists(self):
    # This is kind of a hack. There's no guarantee that a previous
    # RECORDED selection action indicates a selection (or that the
    # lack of one indicates no selection).
    if self.lastaction == 'select':
      return 1

    # Check for selection in the text widget
    try:
      sel = self.text.get('sel.first', 'sel.last')
      return 1
    except TclError:
      return 0

  def extend_selection_command(self, event):
    self.extend_selection()
    
  def extend_selection(self):
    if not self._collect: return
    self.add_action('select', 'select')
  
  def copy_selection(self, event = None):
    if not self._collect: return
    self.add_action('copy', 'copy')
      
  def insert_string(self, event):
    #print 'Inserting:', event.keysym
    if not self._collect: return

    if self.selection_exists():
      # The user typed over the selection: emit a deleteselection
      # action.

      # Set the old state aside to use it as the first state for
      # the insert action (problems with two actions in one event
      # handler).
      last_state = self.last_state

      self.add_action(None, 'deleteselection')

      # Recreate the "proper" last state.
      assert(last_state.selection_extent is not None)
      rowCol = mark_split(last_state.selection_extent[0])
      sel_start = textedit.positionFromTuple(last_state.contents,
                        rowCol)
      rowCol = mark_split(last_state.selection_extent[1])
      sel_end = textedit.positionFromTuple(last_state.contents,
                      rowCol)
      self.last_state = copy.copy(last_state)
      self.last_state.contents = last_state.contents[:sel_start] + \
                  last_state.contents[sel_end:]
      self.last_state.pos = sel_start
      self.last_state.selection_extent = ()


    self.add_action('insert', 'insert')
      
  def move_command(self, event):
    if not self._collect: return
#    print 'Move command:', event.keysym
    if not (event.keysym == 'a' or event.keysym == 'Home' or \
         event.keysym == 'b' or event.keysym == 'Left' or \
         event.keysym == 'e' or event.keysym == 'End' or \
         event.keysym == 'f' or event.keysym == 'Right' or \
         event.keysym == 'n' or event.keysym == 'Down' or \
         event.keysym == 'p' or event.keysym == 'Up' or \
         event.keysym == 'Next' or event.keysym == 'Prior'):
      self.set_status('Warning: key combination %s has no effect' %
        event.keysym)
      return

    self.add_action('move', 'move')
      
  def delete_command(self, event):
    # TODO: think about Delete() vs Cut()
    # Two operations: deleting selection, copying selection to
    # clipboard.  Cut does both, Copy does the latter; it should be
    # possible to do only the former.
    if not self._collect: return

    # Special case: backspace during an insert
    # If the user pressed backspace
    # and we're in an insert
    # and we've inserted more than we've deleted
    current_pos = len(self.text.get('1.0', 'insert'))
    if (event.keysym == 'BackSpace' or event.keysym == 'h') and\
      self.lastaction == 'insert' and\
      self.first_state.pos <= current_pos:
      if self.first_state.pos == current_pos:
        # back out the insert action entirely
        self.lastaction = None
        self.last_state = self.first_state
      else:
        # ignore this action
        pass
      return

    delete_selection_keys = ['x', 'BackSpace', 'Delete']
    if event.keysym in delete_selection_keys:
      if self.selection_exists():
        #self.add_action('cut', 'cut')
        if event.keysym == 'x': # PERCY: cut is cut, deleteselection is deleteselection
          self.add_action('cut', 'cut')
        else:
          self.add_action('deleteselection', 'deleteselection')
        return
      elif event.keysym == 'x':
        return

    if event.keysym == 'd' or event.keysym == 'Delete' or \
       event.keysym == 'h' or event.keysym == 'BackSpace' or \
       event.keysym == 'k':
      self.add_action('delete', 'delete')
    elif event.keysym == 'x':
      return
    else:
      self.set_status('Unsupported delete command: %s' %
        event.keysym)
        
  def paste_command(self, event):
    if not self._collect: return
    #print 'Insert command:', event.keysym
    try:
      selection = self.text.selection_get(selection='CLIPBOARD')
    except TclError:
      # TODO
      self.set_status('Oops!  There is no selection!')
      return

    self.add_action(None, 'paste')
    return 'break'
      
  def cut_command(self, event):
    if not self._collect: return
    if self.selection_exists(): 
      self.add_action('cut', 'cut')
        
  def transpose_command(self, event):
    # TODO: Not currently handled
    if not self._collect: return
    self.set_status('Transpose command "%s" not supported!' %
           event.keysym)
      
  def update_modifiers(self):
    for key in self.__modifiers.keys():
      var = eval('self.%s_var' % key)
      var.set(self.__modifiers[key])
        
  def get_modifier_string(self):
    s = ''
    for mod in MODIFIERS:
      if self.__modifiers[mod]:
        s = s + MOD_MAPPING[mod] + '-'
    return s
  
  def collect_button_press(self, event):
    self.button_down = 1
    if not self._collect: return

    # This should only cause a move if the action actually causes
    # the cursor to move (as it might *not* in a mouse selection that
    # starts where the cursor already is).
    if textedit.rowColTuple(self.last_state.contents,
                self.last_state.pos) != \
                mark_split(self.text.index('insert')):
      self.add_action('move', 'move')
    else:
      #Debug!!!
      print "Skipping move report since the move didn't go anywhere."
        
  def collect_mouse_move(self, event):
    if self.button_down:
      # The user is creating or extending a selection
      self.extend_selection(event)
        
  def collect_key_press(self, event):    
    if event.keysym in MODIFIERS:
      self.__modifiers[event.keysym] = 1
      self.update_modifiers()
      return
    
  def collect_key_release(self, event):
    if event.keysym in MODIFIERS:
      self.__modifiers[event.keysym] = 0
      self.update_modifiers()
      return
    
  def collect_button_release(self, event):
    self.button_down = 0
    if not self._collect: return

    # If there is a selection, then we used the mouse to select
    if self.selection_exists():
      self.extend_selection()


  # ------------------------------------------------------------
  # Scenario event handling
    
  def scenario_key_press(self, event):
    print "scenario_key_press", event
    if event.keysym in MODIFIERS:
      self.__modifiers[event.keysym] = 1
      self.update_modifiers()
      return
    # Record event for scenario playback
    # Ignore alt-keypress
    if self.__modifiers['Alt_R'] or self.__modifiers['Alt_L']:
      return
    char = event.keysym
    print "scenario_key_press", char
    modstring = self.get_modifier_string()
    now = time.time()
    e = Event('<%sKey-%s>' % (modstring, char),
      now - self.scenario_time)
    self.scenario_time = now
    self.scenario_events.append(e)
  
  def scenario_key_release(self, event):
    if event.keysym in MODIFIERS:
      self.__modifiers[event.keysym] = 0
      self.update_modifiers()
      return

  def scenario_button_press(self, event):
    now = time.time()
    e = Event('<ButtonPress-%i>' % event.num,
      now - self.scenario_time)
    self.scenario_time = now
    e['pos'] = tkTextClosestGap(self.text, event.x, event.y)
    self.scenario_events.append(e)

  def scenario_button_release(self, event):
    now = time.time()
    e = Event('<ButtonRelease-%i>' % event.num,
      now - self.scenario_time)
    self.scenario_time = now
    e['pos'] = tkTextClosestGap(self.text, event.x, event.y)
    self.scenario_events.append(e)

  def scenario_mouse_move(self, event):
    if not event.state:
      # Record only motion when a button is pressed
      return
    now = time.time()
    e = Event('<Motion>', now - self.scenario_time)
    self.scenario_time = now
    e['pos'] = tkTextClosestGap(self.text, event.x, event.y)
    e['state'] = event.state
    self.scenario_events.append(e)
  # ------------------------------------------------------------

  # ------------------------------------------------------------
  # Event counters
  def count_button_press(self, event):
    self.mouseclicks = self.mouseclicks + 1
  def count_key_press(self, event):
    self.keypresses = self.keypresses + 1
    
  def reset_counters(self):
    self.mouseclicks = 0
    self.keypresses = 0
  
  def show_counters_popup(self):
    m = tkMessageBox.askyesno(title='Event counters',
      message='%i keypresses\n%i mouse clicks\n\nReset counters?' %
      (self.keypresses, self.mouseclicks))
    if m:
      self.reset_counters()
  
  def show_counters_text(self):
    print '%i keypresses\n%i mouse clicks' % (
      self.keypresses, self.mouseclicks)
  # ------------------------------------------------------------

  def show_hypothesis(self):
    if not self.vs:
      tkMessageBox.showerror('What?',
        'Please teach me something first.')
      return
    reasons = self.vs.reasons()
    text = string.join(reasons, '\n')
    t = ShowHypothesis('Best hypothesis',
      lambda s=self: string.join(s.vs.reasons(), '\n'))
  
  def show_priors(self):
    if not self.vs:
      tkMessageBox.showerror('What?',
        'Please teach me something first.')
      return
    p = self.vs.getpriors()
    s = vs.format_priors(p)
    t = ShowHypothesis('Prior probabilities', lambda s=self:
      vs.format_priors(s.vs.getpriors()))

  def collecting(self):
    return self._collect

  # Training the version space
  def start_collect(self):
    if self._collect:
      return
    
    if self.undostate:
      self.loadundostate()
      self.undostate = None
    self.trackingundostate = self.getundostate()
    self.lastundostate = self.trackingundostate

    self.current_action = None
    if SCREEN_SHOT_MODE:
      self.macrobut.config(text='Stop\nrecording',
               command=self.stop_collect,
               background='maroon')
    else:
      self.macrobut.config(text='Stop recording',
               command=self.stop_collect,
               background='maroon')

    self.first_state = self.get_text_state('insert')
    self.last_state = self.first_state
    self.change_list = []
    self.lastaction = None

    self.set_status('Started recording')
    self._collect = 1

  def system_stop_collect(self):
    if not self._collect:
      return
    
    self.set_status('Finished recording')
    self.add_action(None, None)

    self._collect = 0
    oldbg = self.macrobut.configure('background')[3]
    if SCREEN_SHOT_MODE:
      self.macrobut.config(text='Record\niteration',
        command=self.start_collect, background=oldbg)
    else:
      self.macrobut.config(text='Record iteration',
        command=self.start_collect, background=oldbg)

  def cancel_collect(self):
    # Kill the current collect *WITHOUT* recording the current
    # action.
    self.lastaction = None
    self.loadundostate(self.lastundostate)
    self.system_stop_collect()

  def stop_collect(self):
    if not self._collect:
      return

    self.system_stop_collect()
    self.vs.feed_end()

    if self.active_learning:
      #DEBUG!!
      #print 'Advising user.'
      self.advise_user()
  
  def is_recording(self):
    return self._collect

  def no_more_examples(self):
    # There are no more examples to do

    # For the moment, I'm also going to invalidate the last action
    # recorded.
    if self._collect and self.lastaction == 'move':
      self.lastaction = None
    
    self.vs.feed_terminate(self.get_text_state('insert'))

  class UndoData:
    fields_to_record = ['_collect',
            'lastundostate',
            'lastaction',
            'trackingundostate',
            'current_action',
            'online_learning',
            'undostate',
            'last_state',
            'first_state',
            ]
    def __init__(self, app):
      for field in self.fields_to_record:
        assert(hasattr(app, field))
        setattr(self, field, getattr(app, field))
      self.__undostate = app.getundostate()

    def restore(self, app):
      for field in self.fields_to_record:
        assert(hasattr(app, field))
        assert(hasattr(self, field))
        setattr(app, field, getattr(self, field))
      app.loadundostate(self.__undostate)

  # Backup one step
  def undo(self):
    if not self.undo_trace:
      self.set_status("No further undo information!")
    else:
      (program, state) = self.undo_trace[0]
      undo_trace = self.undo_trace[1:]

      state.restore(self)
      self.load_program(program)
      self.undo_trace = undo_trace
      self.set_status("Restored previous state.")

  # Set aside the state for undoing; DOES NOT commit that state to the
  # undo list.
  def record_undo_step(self):
    self.current_undo_step = self.UndoData(self)

  # Record a step for later undoing
  def save_undo_step(self):
    program = self.get_program()
    self.undo_trace = [(program, self.current_undo_step)] + \
             self.undo_trace[:self.max_undo_steps - 1]

  def advice_string(self, difficulty,
           default = "Oops, something bad happened!"):
    # Each entry shows a message to give for difficulties above the
    # given level.
    ADVICE = [(0, "I believe I have the idea!"),
        (1.0, "You should step through this."),
        (2.0, "Maybe you should step through this."),
        (3.0, "I suggest you perform another example."),
        (4.0, "Who, what, when, where, why?")]
    

    list = filter(lambda i, diff = difficulty: i[0] <= diff,
            ADVICE)
    if list:
      return list[-1][1]
    else:
      return default
    
  def advise_user(self):
    """Suggest to the user whether to perform another example."""

    try:
      if not self.vs or self.vs.action_count() == 0:
        return
    except vs.VSError:
      return

    # Until self.vs.action_count() actually reflects something like
    # the most probable action count (or at least isn't normally 1!)
    # I'm going to use an arbitrary number.
    diff = self.rate_difficulty()
    str = self.advice_string(diff)
    self.set_status("%s (difficulty is %s)" % (str, diff))

  def generate_choices(self, num_per_choice_type):
    import ent_util
    state = self.get_text_state('insert')
    generators = [
      ent_util.RandomExamplePicker(self.vs, state),
      ent_util.LinewiseExamplePicker(self.vs, state),
      ent_util.MoveBasedExamplePicker(self.vs, state),
      ]

    # Create a list of num_per.. sets of points from each generator,
    # throwing out points that come back as 'None'.
    examples = filter(None, map(lambda x: x.next_example(),
                 generators * num_per_choice_type))

    # Sort and remove duplicates.
    examples.sort(lambda x, y: cmp(textedit.mark_split(x),
                 textedit.mark_split(y)))
    i = 1
    while i < len(examples):
      if examples[i] == examples[i-1]:
        del examples[i]
      else:
        i = i + 1

    return examples
    
    
  def jump(self, num_per = 5, tolerance = None):
    """Jump to new example and record that example.

    Returns to current state on stop record but retains learned
    information. Attempts to pick best available example.
    """

    self.set_status("Jumping to a new example. " + \
           "Record on that then hit stop to return.")
    self.text.update_idletasks()

    undostate = self.getundostate()

    # Pick a swath of points to consider
    choices = self.generate_choices(num_per)

    # Select the point to jump to.
    (choice, calc) = self.select_point(choices, tolerance)
    self.text.mark_set('insert', choices[choice])
    self.text.see('insert')

    self.start_collect()

    def wrap_stop_collect(s=self, u=undostate,
              num=self.vs.action_number):
      s.stop_collect()
      s.loadundostate(u)
      #HACK!!!
      # Need to reset the program counter here.
      s.vs.action_number = num
    
    self.macrobut.config(command = wrap_stop_collect)

  def get_entropy_calculator(self, location, length):
    action_num = self.vs.get_action_number()

    # Calculate log leaf counts:
    branch_counts = map(self.vs.num_hyps_nth,
            range(action_num, action_num + length))
    log_branch_counts = map(entropy.ent_log, branch_counts)
    # Initialize counts to 1 with extra slot for convenience.
    log_leaf_counts = [0] * (len(log_branch_counts) + 1)
    for i in range(len(log_branch_counts), 0, -1):
      log_leaf_counts[i - 1] = log_leaf_counts[i] + \
                log_branch_counts[i - 1]
    del log_leaf_counts[-1]

    executor = lambda i, s, b=action_num, v=self.vs: v.execute_nth(i,s+b)
    return entropy.EntropyCalculator(log_leaf_counts,
                   length,
                   executor,
                   self.simulate_action,
                   self.get_text_state(location))

  def select_point(self, locations, tolerance = None):
    """Select the point with the highest entropy of the given points."""
    action_num = self.vs.get_action_number()
    action_count = self.vs.action_count()

    # Calculate log leaf counts:
    branch_counts = map(self.vs.num_hyps_nth,
            range(action_num, action_num +
               self.vs.action_count()))
    log_branch_counts = map(entropy.ent_log, branch_counts)
    # Initialize counts to 1 with extra slot for convenience.
    log_leaf_counts = [0] * (len(log_branch_counts) + 1)
    for i in range(len(log_branch_counts), 0, -1):
      log_leaf_counts[i - 1] = log_leaf_counts[i] + \
                log_branch_counts[i - 1]
    del log_leaf_counts[-1]

    num = len(locations)
    executor = lambda i, s, b=action_num, v=self.vs: v.execute_nth(i,s+b)
    ecs = map(entropy.EntropyCalculator,
        [log_leaf_counts] * num,
        [action_count] * num,
        [executor] * num,
        [self.simulate_action] * num,
        map(self.get_text_state, locations))

    if tolerance is None:
      tolerance = 0.5
    choice = entropy.choose_highest_entropy(ecs, tolerance)
    return (choice, ecs[choice])

  def rate_difficulty(self, length = None, tolerance = 0.5,
          time_limit = 0.5):
    """Rate the difficulty of guessing the macro in the current state."""

    #DEBUG!!
    #print 'Final probs:', probs
    action_num = self.vs.get_action_number()

    # Calculate log leaf counts:
    branch_counts = map(self.vs.num_hyps_nth,
            range(action_num, action_num +
               self.vs.action_count()))
    log_branch_counts = map(entropy.ent_log, branch_counts)
    # Initialize counts to 1 with extra slot for convenience.
    log_leaf_counts = [0] * (len(log_branch_counts) + 1)
    for i in range(len(log_branch_counts), 0, -1):
      log_leaf_counts[i - 1] = log_leaf_counts[i] + \
                log_branch_counts[i - 1]
    del log_leaf_counts[-1]
    #DEBUG!!
    #print 'Log leaf counts:', log_leaf_counts
    ec = entropy.EntropyCalculator(log_leaf_counts,
                 self.vs.action_count(),
                 # Construct an executor which takes
                 # steps numbered from 0.
                 lambda i, s, b=action_num, v=self.vs:
                 v.execute_nth(i, s+b),
                 self.simulate_action,
                 self.get_text_state('insert'))

    ec.timed_approx_calc(time_limit = time_limit,
             tolerance_percent = tolerance)
    return ec.get_entropy_estimate()

  def abort(self):
    self._collect = 0
    oldbg = self.macrobut.configure('background')[3]
    self.macrobut.config(text='Record iteration',
      command=self.start_collect, background=oldbg)
    self.clear_events()

  # Guess button
  def guess_again(self):
#    self.last_rank = (self.last_rank + 1) % len(self.current_actions)
    self.loadundostate()
    (hyplist, totalprob) = self.result_set.pick_next_best()
    self.perform_action(hyplist, totalprob, 0)
#    totalvotes = reduce(operator.add,
#      map(lambda x: x[0], self.current_actions))
#    self.perform_action(self.current_actions[self.last_rank],
#      totalvotes)
  def enable_guess_button(self):
    self.guessbut.config(state='normal')
  def disable_guess_button(self):
    self.guessbut.config(state='disabled')

  def write_program(self, filename, program):
    f = open(filename, 'wb')
    pickle.dump(program, f)

  def read_program(self, filename):
    f = open(filename, 'rb')
    return pickle.load(f)

  def get_program(self):
    return copy.deepcopy(self.vs)

  def save_program(self):
    self.saved_program = self.get_program()

  def load_program(self, program = None):
    if program is None:
      program = self.saved_program

    if program is None:
      raise 'no program to load'
    else:
      self.vs = program

  def saveundostate(self):
    self.undostate = self.getundostate()

  def getundostate(self):
    text = self.text.get('1.0', 'end - 1c')
    view = (self.text.xview(), self.text.yview())
    cursorpos = self.text.index('insert')
    try:
      selstart = self.text.index('sel.first')
      selend = self.text.index('sel.last')
    except TclError:
      selstart = None
      selend = None
    return (text, cursorpos, view, selstart, selend)
  
  def loadundostate(self, loadstate = None):
    if loadstate is None:
      loadstate = self.undostate
    
    try:
      text, cursorpos, view, startsel, endsel = loadstate
    except:
      traceback.print_exc()
      return
    self.text.delete('1.0', 'end')
    self.text.insert('end', text)
    self.text.xview('moveto', view[0][0])
    self.text.yview('moveto', view[1][0])
    self.text.mark_set('insert', cursorpos)
    self.text.tag_remove('sel', '1.0')
    if startsel is not None and endsel is not None:
      self.text.tag_add('sel', startsel, endsel)

  # Simulate an action on the given input state. Return the resulting
  # output state.
  def simulate_action(self, hypothesis, instate):
    # DEBUG!!
    #print 'Simulating:', hypothesis
    (command, arg, callback) = hypothesis

    if vs.hyp_inapplicable(arg):
      return None
    
    outstate = copy.copy(instate)

    if command == 'move':
      newpos = textedit.positionFromTuple(outstate.contents, arg)
      outstate.pos = newpos
      outstate.selection_extent = ()
    elif command == 'select':
      sel_end = textedit.positionFromTuple(outstate.contents, arg[0])
      newpos = textedit.positionFromTuple(outstate.contents, arg[1])
      oldtuple = textedit.rowColTuple(outstate.contents, outstate.pos)
      if sel_end > instate.pos:
        outstate.selection_extent = (mark_create(oldtuple),
                      mark_create(arg[0]))
      else:
        outstate.selection_extent = (mark_create(arg[0]),
                      mark_create(oldtuple))
      outstate.pos = newpos
    elif command == 'delete':
      fore_pos = textedit.positionFromTuple(outstate.contents, arg[1])
      back_pos = textedit.positionFromTuple(outstate.contents, arg[0])
      
      outstate.contents = outstate.contents[:back_pos] + \
              outstate.contents[fore_pos:]
      outstate.pos = back_pos
      outstate.selection_extent = ()
    elif command == 'insert':
      insert_string = arg

      outstate.contents = outstate.contents[:outstate.pos] + \
              insert_string + \
              outstate.contents[outstate.pos:]
      outstate.pos = outstate.pos + len(insert_string)
      outstate.selection_extent = ()      
    elif command == 'cut':
      back_tuple = mark_split(outstate.selection_extent[0])
      fore_tuple = mark_split(outstate.selection_extent[1])
      fore_pos = textedit.positionFromTuple(outstate.contents,
                       fore_tuple)
      back_pos = textedit.positionFromTuple(outstate.contents,
                       back_tuple)
      outstate.clipboard = outstate.contents[back_pos:fore_pos]
      outstate.contents = outstate.contents[:back_pos] + \
              outstate.contents[fore_pos:]
      outstate.pos = back_pos
      outstate.selection_extent = ()      
    elif command == 'copy':
      back_tuple = mark_split(outstate.selection_extent[0])
      fore_tuple = mark_split(outstate.selection_extent[1])
      fore_pos = textedit.positionFromTuple(outstate.contents,
                       fore_tuple)
      back_pos = textedit.positionFromTuple(outstate.contents,
                       back_tuple)
      outstate.clipboard = outstate.contents[back_pos:fore_pos]
      outstate.selection_extent = ()      
    elif command == 'paste':
      insert_string = outstate.clipboard

      outstate.contents = outstate.contents[:outstate.pos] + \
              insert_string + \
              outstate.contents[outstate.pos:]
      outstate.pos = outstate.pos + len(insert_string)
      outstate.selection_extent = ()      
    elif command == 'deleteselection':
      back_tuple = mark_split(outstate.selection_extent[0])
      fore_tuple = mark_split(outstate.selection_extent[1])
      fore_pos = textedit.positionFromTuple(outstate.contents,
                       fore_tuple)
      back_pos = textedit.positionFromTuple(outstate.contents,
                       back_tuple)
      outstate.contents = outstate.contents[:back_pos] + \
              outstate.contents[fore_pos:]
      outstate.pos = back_pos
      outstate.selection_extent = ()

    return outstate

  # Perform an action.  Hypothesis should be a vs.Hypothesis instance.
  def perform_action(self, hyplist, totalprob, sole_output_state = 0):
    hypothesis = hyplist[0][0]
    (command, arg, callback) = hypothesis.hyp
    if DEBUG: print 'Hyp.hyp:', hypothesis.hyp
    if self.online_learning:
      self.first_state = self.get_text_state('insert')
      
    if arg is None:
      argstr = ' '
    else:
      argstr = ' ' + repr(arg)

    ## Make sure we don't attribute 0% probability to anything with
    ## non-zero probability and don't attribute 100% probability
    ## if prob < 100%
    probability = totalprob*100
    if not sole_output_state:
      if probability > 99:
        probability = 99
      if probability < 1:
        probability = 1
    self.set_status('%s [%2.f%% prob.]' %
           (hypothesis.reason, probability))
    self.current_action = (command, arg, callback)
    self.saveundostate()

    if vs.hyp_inapplicable(arg):
      self.set_status('Terminate due to: %s [%2.f%% prob.]' %
             (hypothesis.reason, probability))
    elif command == 'move':
      loc = '%i.%i' % (arg[0], arg[1])
      self.text.tag_remove('sel', '1.0', 'end')
      self.text.mark_set('insert', loc)
      self.text.see(loc)
    elif command == 'select':
      loc = '%i.%i' % (arg[1][0], arg[1][1])
      sel_end = '%i.%i' % (arg[0][0], arg[0][1])
      self.text.tag_remove('sel', '1.0', 'end')
      here = self.text.index('insert')
      if mark_greater(sel_end, here):
        self.text.tag_add('sel', here, sel_end)
      else:
        self.text.tag_add('sel', sel_end, here)
      self.text.mark_set('insert', loc)
      self.text.see(loc)
    elif command == 'delete':
      loc_back = '%i.%i' % (arg[0][0], arg[0][1])
      loc_fore = '%i.%i' % (arg[1][0], arg[1][1])
      self.text.tag_configure('overstrike', overstrike=1,
        foreground='white', background='black')
      self.text.tag_add('overstrike', loc_back, loc_fore)
      self.text.mark_set('insert', loc_back)
      self.text.see('insert')
    elif command == 'insert':
      self.text.insert('insert', arg)
      self.text.see('insert')
    elif command == 'cut':
      self.text.event_generate('<<Cut>>')
    elif command == 'copy':
      self.text.event_generate('<<Copy>>')
    elif command == 'paste':
      self.text.event_generate('<<Paste>>')
    elif command == 'deleteselection':
      try:
        self.text.delete('sel.first', 'sel.last')
      except TclError:
        self.set_status('Error deleting non-existent selection')
      self.text.see('insert')

  def post_execute(self, action):
    self.undostate = None
    
    (command, arg, callback) = action
    if vs.hyp_inapplicable(arg):
      raise EndProgram, 'prediction of inapplicable executed'
    elif command == 'delete':
      # Construct the extents of the delete
      loc_back = '%i.%i' % (arg[0][0], arg[0][1])
      loc_fore = '%i.%i' % (arg[1][0], arg[1][1])

      # Perform the delete
      self.text.delete(loc_back, loc_fore)
      self.text.see(loc_back)

    if callback is not None:
      if DEBUG: print 'Calling callback', callback
      callback()
    
    # Record this action as an example.
    if self.online_learning:
      # self.record_undo_step() # UNDO IS NOT WORKING
      self.create_action(command, self.first_state,
                self.get_text_state('insert'))
    else:
      pass

    # Code for using simulation to post_execute:
    #
    #self.loadundostate()
    #outstate = self.simulate_action(action,
    #              self.get_text_state('insert'))
    #
    #self.set_text_state(outstate)

  def finish_step(self, action = None):
    if action is None:
      action = self.current_action
      
    if not action:
      print 'Finish_step called with no step to finish!'
      return
    
    # Do the next step
    try:
      self.post_execute(self.current_action)
    finally:
      self.current_action = None

  def step_vector(self):
    if self._collect:
      self.system_stop_collect()
    
    # OK, how do we actually perform the actions we've learned?
    if not self.vs:
      self.set_status('No macro learned!')
      return

    if self.current_action:
      try:
        self.finish_step(self.current_action)
      except EndProgram:
        self.set_status('Ending program.')
        self.disable_guess_button()
        return 1

    self.show_step()

    return 0

  def show_step(self):
    self.result_set = self.vs.execute_nth(
      self.get_text_state('insert'))

    # PERCY
    iter = self.vs.action_number
    numIters = self.vs.action_count()
    #print "ITER", iter, numIters
    if iter % numIters == 0:
      self.addHighLevelAction(None)

    #if DEBUG: print "show_step: result_set = ", self.result_set
    #if not self.result_set:
    if empty(self.result_set): # PERCY
      # No more predictions
      self.set_status('Error: no more predictions!')
      self.disable_guess_button()
      return 0

    self.result_set.initialize()
    (hyplist, totalprob) = self.result_set.pick_next_best()
    # TODO: do something with the list.  For now we'll just pick the
    # most likely hypothesis off the list.
    # Here's the list of reasons:
    #reasons = map(lambda x: x[0].reason, hyplist)

    self.perform_action(hyplist, totalprob,
            self.result_set.num_distinct_actions == 1)

    # Enable the guess button if there is more than one disinct action
    if self.result_set.num_distinct_actions > 1:
      self.enable_guess_button()
    else:
      self.disable_guess_button()

    return 1

  #DEBUG!!!
  # Entropy testing code
  def entropy_choice(self):
    import ent_util
    ent_util.ExampleChoicePopup(self)

  # ----------------------------------------------------------------------
  # Scenario functions

  def open_scenario(self):
    self.edit_scenario()
  
  def edit_scenario(self):
    db.ScenarioPopup(self)
  
  def iterate_scenario(self, scenario, exnum=None):
    newstate = self.get_text_state('insert')
    scenario.add_iteration(self.scenario_state,
      self.scenario_events, newstate, exnum)
    self.scenario_state = newstate
    self.scenario_events = []

  def record_scenario(self):
    # Hack to get around the lack of keyrelease events when you
    # invoke a menu with the alt key
    for key in self.__modifiers.keys():
      self.__modifiers[key] = 0
    self.update_modifiers()

    # Scenario recording state variables
    self.scenario_state = self.get_text_state('insert')
    self.scenario_events = []
    self.set_status('Recording scenario')
    self.scenario_time = time.time()
    # Switch over to scenario-recording keybindings
    self.bindkeys_scenario()
  
  def done_scenario(self, scenario):
    self.iterate_scenario(scenario)
    self.set_status('Done recording scenario')
    self.bindkeys_macro()
  
  def record_single_iteration(self, example):
    self.set_text_state(example.beforestate)
    self.record_scenario()
  
  def done_single_iteration(self, scenario, exnum):
    self.iterate_scenario(scenario, exnum)
    self.set_status('Done recording scenario')
    self.bindkeys_macro()
  
  def replay_scenario_examples(self, scenario, ex_list):
    if not ex_list:
      return

    self.set_status('Playing back example #%i' % (ex_list[0] + 1))
    example = scenario[ex_list[0]]
    events = example.events
    state = example.beforestate
    self.text.delete('1.0', 'end')
    self.text.insert('end', state.contents)
    self.text.mark_set('insert',
              mark_create(textedit.rowColTuple(state.contents,
                          state.pos)))
    self.text.see('insert')

    if not events: return

    self.text.grab_set()
    self.text.focus_set()
    timediff = events[0].timedelta
    self.scenario_proc = self.text.after(int(timediff * 250),
      lambda e=events, s=self, sc=scenario, l=ex_list:
        s.run2(e, sc, l))

  def run2(self, events, scenario, ex_list):
    assert(events)
    event = events[0]
    self.set_status(str(event))
    event.generate(self.text)
    events = events[1:]
    if not events:
      self.text.grab_release()
      self.set_status('Done.')
      donestate = self.get_text_state('insert')
      assert(donestate == scenario[ex_list[0]].afterstate)

      # Start next example
      if ex_list:
        self.replay_scenario_examples(scenario, ex_list[1:])
    else:
      self.scenario_proc = self.text.after(
        int(events[0].timedelta * 250),
        lambda e=events, s=self, sc=scenario, l=ex_list:
        s.run2(e, sc, l))

  # This doesn't work
  def cancel_run(self):
    if self.scenario_proc:
      self.after_cancel(self.scenario_proc)
      self.text.grab_release()
      self.set_state('Playback cancelled.')
      self.scenario_proc = None
  # ----------------------------------------------------------------------

  # ----------------------------------------------------------------------
  # Training and testing
  def run_example(self, state, events):
    self.set_text_state(state)
    self.text.focus_set()
    self.text.grab_set_global()
    for e in events:
      e.generate(self.text)
    self.text.grab_release()
    return self.get_text_state('insert')
  
  def collect_example(self, state, events, iteration):
    self.set_text_state(state)
    self.text.focus_set()
    self.text.grab_set_global()

    if DEBUG:
      print 'Starting to collect iteration #', iteration
    self.start_collect()
    for e in events:
      #if DEBUG: print 'modifiers:', self.__modifiers
      if DEBUG: print 'Generating event', e
      e.generate(self.text)
    self.stop_collect()
    self.text.grab_release()
    return self.get_text_state('insert')
  
  def try_exec_vs(self, beforestate, afterstate, iteration):
    self.set_text_state(beforestate)
    self.text.focus_set()
    self.text.grab_set_global()
    donestate = None
    try:
      num_actions = self.vs.action_count()
      base = num_actions * iteration
      self.action_number = base
      for i in range(num_actions):
        # Pick most likely hypothesis
        result_set = self.vs.execute_nth(
          self.get_text_state('insert'))
        #if not result_set:
        if empty(result_set): # PERCY
          # No guesses; we failed
          return 0
        result_set.initialize()
        (hyplist, totalprob) = result_set.pick_next_best()
        # Choose the first hypothesis in the list
        self.perform_action(hyplist, totalprob)
        try:
          self.post_execute(self.current_action)
        except EndProgram:
          # Don't know what to do about this yet!
          # BUG!!!
          break
      donestate = self.get_text_state('insert')
    finally:
      self.text.grab_release()
    return donestate == afterstate
  
  def test_scenario(self, scenario, interactive=1):
    self.filename = None
    numexs = len(scenario)
    if DEBUG: print 'at beginning of test_scenario'
    #for i in range(numexs-1):
    for i in range(numexs): # PERCY: don't print -1 even if require all training examples
      if DEBUG: print 'outer loop iteration var', i
      self.text.update_idletasks()
      # Try a training/test split on example i
      # Train on examples[:i+1], and test on examples[i+1:]
      training_exs = scenario[:i+1]
      test_exs = scenario[i+1:]
      self.reset()

      iter = 0

      # Train
      for ex in training_exs:
        if DEBUG:
          print 'collecting examples in inner loop, iter', iter
        afterstate = self.collect_example(ex.beforestate,
          ex.events, iter)
        self.text.update_idletasks()
        if afterstate != ex.afterstate:
          print 'Error: bad training example!'
          print ex.afterstate
          print afterstate
          out = open('a1', 'w')
          out.write(ex.afterstate.contents)
          out.close()
          out = open('a2', 'w')
          out.write(afterstate.contents)
          out.close()
          return (-2, numexs) # PERCY
          #return (-1, numexs)
        iter = iter + 1

      break # TMP

      # Test
      for ex in test_exs:
        result = self.try_exec_vs(ex.beforestate,
          ex.afterstate, iter)
        self.text.update_idletasks()
        if not result:
          # Don't bother testing on the remainder of the
          # training examples
          # print 'Test example failed'
          break
        else:
          # It was correct
          pass
        iter = iter + 1
      else:
        # We performed correctly on this example
        if (i+1)==1:
          s = ''
        else:
          s = 's'
        out = 'Testing successful with %i training' % (i+1)
        out = out + ' example%s (%i total)' % (s, numexs)
        if interactive:
          tkMessageBox.showinfo('Test results', out)
        return (i+1, numexs)
    if interactive:
      tkMessageBox.showerror('Learning failed',
        'Learning unsuccessful (%i examples total)' % numexs)
    return (-1, numexs)
  
  def test_all_scenarios(self):
    s = '%-30s%10s%10s\n' % ('Scenario', '# train', '# total')
    s = s + ('-' * 50) + '\n'
    scenario_list = db.get_scenarios()
    scenario_list.sort()
    for scenario_name in scenario_list:
      if scenario_name != 'number-iterations': continue # TMP
      print 'Testing scenario', scenario_name
      scenario = db.load_scenario(scenario_name)
      (exs, total) = self.test_scenario(scenario, 0)
      s = s + '%-30s%10i%10i\n' % (scenario_name, exs, total)
    d = Pmw.TextDialog(self, title='Scenario results')
    d.insert('end', s)

  def dump_all_scenarios(self):
    # Convert pickled object format to YAML format
    outPath = "yaml"
    os.system("mkdir -p "+outPath)

    scenario_list = db.get_scenarios()
    scenario_list.sort()
    numGood = 0
    for scenario_name in scenario_list:
      #if scenario_name != 'command-line-args': continue # TMP
      print 'Dumping scenario', scenario_name
      scenario = db.load_scenario(scenario_name)
      self.reset()

      map = {}
      map['name'] = scenario.name
      map['description'] = scenario.description
      map['examples'] = []

      iter = 0
      for ex in scenario.examples: # For each example, do it
        #print iter, ex.events
        self.high_level_actions = []
        afterstate = self.collect_example(ex.beforestate, ex.events, iter)
        self.text.update_idletasks()
        if afterstate != ex.afterstate:
          print 'ERROR: bad training example!'
          break
        iter += 1
        exmap = {}
        exmap['before'] = self.convertState(ex.beforestate)
        exmap['after'] = self.convertState(ex.afterstate)
        # Look at the actions that resulted from that example
        actions = []
        try:
          for action, instate, outstate in self.high_level_actions:
            actions.append(self.convertAction(action, instate, outstate))
        except:
          print 'ERROR: crappy inconsistency with fine-grained simulation'
          break
        print " ", repr(" | ".join(actions))
        exmap['actions'] = actions
        map['examples'].append(exmap)
      else:
        # Output the scenario if all its examples are good
        numGood += 1
        self.writeScenarioYamlExamples(outPath+"/"+name, map)
    print numGood, "out of", len(scenario_list), "scenarios were valid and dumped"

  def writeScenarioYamlExamples(self, pathPrefix, map):
    name = map['name']
    examples = map['examples']
    assert (len(examples) > 0)
    f = open(pathPrefix+".yaml", "w")
    f.write(yaml.dump(map))
    f.close
    exPath = pathPrefix+".examples"
    os.system("mkdir -p "+exPath)
    for i, ex in zip(xrange(len(examples)), examples):
      f = open("%s/%03d" % (exPath, i), "w")
      f.write(ex['before']['contents'])
      f.close
    f = open("%s/%03d" % (exPath, len(examples)), "w")
    f.write(examples[len(examples)-1]['after']['contents'])
    f.close
    print name+": %d examples" % len(examples)

  def convertAction(self, action, instate, outstate):
    #print "ACT", action
    # All arguments involve absolute position
    if action == 'move':
      e = "%s %s" % (action, outstate.pos)
    elif action == 'select':
      assert (len(outstate.selection_extent) > 0) # Selected something
      e = "%s %s" % (action, self.formatSelection(outstate))
    elif action == 'paste':
      #print len(instate.contents)
      #print len(outstate.contents)
      #print repr(instate.contents)
      #print repr(outstate.contents)
      #print repr(outstate.clipboard)
      n = len(outstate.contents) - len(instate.contents)
      assert (n == len(instate.clipboard))
      assert (n == len(outstate.clipboard))
      e = action
    elif action == 'cut':
      assert (len(instate.selection_extent) > 0) # Cut something
      #print len(instate.contents)
      #print len(outstate.contents)
      #print repr(outstate.clipboard)
      assert (len(instate.contents) - len(outstate.contents) == len(outstate.clipboard)) # Thing cut goes to clipboard
      e = action
    elif action == 'copy':
      assert (len(instate.selection_extent) > 0) # Copied something
      e = action
    elif action == 'delete':
      n = len(instate.contents) - len(outstate.contents)
      if instate.pos == outstate.pos: # Deleting forwards
        e = "%s %s" % (action, instate.pos+n)
      elif instate.pos - n == outstate.pos: # Deleting backwards
        e = "%s %s" % (action, instate.pos-n)
      else:
        assert (False)
    elif action == 'insert':
      n = len(outstate.contents) - len(instate.contents)
      #print "insert", n, len(outstate.contents)
      assert (n > 0)
      assert (instate.pos+n == outstate.pos) # Cursor always advances
      e = "%s %s" % (action, outstate.contents[outstate.pos-n:outstate.pos])
    elif action == 'deleteselection':
      assert (len(instate.selection_extent) > 0) # Deleted something
      e = action
    return e

  def convertState(self, state):
    map = {}
    map['pos'] = state.pos
    map['selection'] = self.formatSelection(state)
    map['clipboard'] = state.clipboard
    map['contents'] = state.contents
    return map

  def formatSelection(self, state):
    se = state.selection_extent
    if len(se) > 0:
      a = textedit.positionFromTuple(state.contents, mark_split(se[0]))
      b = textedit.positionFromTuple(state.contents, mark_split(se[1]))
      return "%d,%d" % (a,b)
    else:
      return None

  # ----------------------------------------------------------------------
  # Interruptable SMARTedit controlled program run.

  # Class to control interrupt button window
  class InterruptUI(Toplevel):
    KEY = 'Escape'
    BOUND_KEY = '<Key-%s>' % KEY
    def __init__(self, master):
      self.master = master
      self.binding = self.master.text.bind(self.BOUND_KEY,
                      lambda event, s=self:
                      s.run_command())
      self.master.set_status('Hit %s to interrupt.' % self.KEY, 0.75)

    def run_command(self):
      self.master.interrupt_run()

    def destroy(self):
      self.master.text.unbind(self.BOUND_KEY, self.binding)
  
  def interruptable_run(self, timeout_tup = None,
            callback = None, max_steps = None):
    # Run actions with a break of "init_timeout" length to allow
    # the user to interrupt. Decrease this timeout at the decay rate
    # to a minimum length of min_timeout. All delays are in seconds.
    #
    # Assumes the action_number is correct.
    #
    # Calls 'callback' with a single argument when it terminates.
    #
    # Performs at most max_steps steps.
    #
    # If the timeout_tup is None, use playback_speed instead.

    assert(max_steps is None or max_steps > 0)
    if callback is None:
      callback = lambda arg: None
    self.interrupt_callback = callback
    self.interrupt_steps = max_steps

    #DEBUG!!
    #print 'Beginning interruptable run.'

    assert(not self._collect)
    if not self.vs:
      self.set_status('No macro learned!')
      return

    if self.undostate:
      self.loadundostate()
      self.undostate = None
    
    # Create a window with the interrupt button and link back
    # to the interrupt method.
    self.intrunbut.config(state='disabled')

    #DEBUG!!
    #print 'Creating interrupt window.'
    self.interrupt_window = self.InterruptUI(self)

    # Begin running the actions.
    self.current_action = None
    self.do_continue_interruptable_run = 1
    self.continue_interruptable_run(timeout_tup)

  def continue_interruptable_run(self, timeout_tup=None):
    # If timeout_tup is None, use playback_speed instead.
    if not self.do_continue_interruptable_run:
      self.current_action = None
      return
    
    # Do the next step
    if self.current_action:
      try:
        self.finish_step(self.current_action)
      except EndProgram:
        self.set_status('Ending program.')
        self.interrupt_run(1)
        return

    if self.interrupt_steps is not None and self.interrupt_steps == 0:
      self.after(0, self.interrupt_run)
      return

    if self.interrupt_steps is not None:
      self.interrupt_steps = self.interrupt_steps - 1

      
    #DEBUG!!
    #print 'Action number:', self.action_number
    self.result_set = self.vs.execute_nth(
      self.get_text_state('insert'))

    # PERCY
    iter = self.vs.action_number
    numIters = self.vs.action_count()
    #print "ITER", iter, numIters
    if iter % numIters == 0:
      self.addHighLevelAction(None)

    #if not self.result_set:
    if empty(self.result_set): # PERCY
      # No more predictions
      self.interrupt_run(1)
      self.set_status('Error: no more predictions!')
      return

    self.result_set.initialize()
    (hyplist, totalprob) = self.result_set.pick_next_best()

    self.perform_action(hyplist, totalprob,
            self.result_set.num_distinct_actions == 1)

    if timeout_tup is None:
      assert hasattr(self, 'playback_speed')
      timeout = self.playback_speed.get()
      self.text.after(int(timeout * 1000),
        lambda s=self: s.continue_interruptable_run())
    else:
      (timeout, decrement, min_timeout) = timeout_tup
      new_timeout = timeout - decrement
      if new_timeout < min_timeout:
        new_timeout = min_timeout
      self.text.after(int(timeout*1000),
        lambda s=self, t=new_timeout, d=decrement,
        m=min_timeout: s.continue_interruptable_run((t,d,m)))

  def interrupt_run(self, terminated = 0):
    if not self.do_continue_interruptable_run:
      return
    self.do_continue_interruptable_run = 0
    if self.undostate:
      self.loadundostate()
      self.undostate = None
    if hasattr(self, 'interrupt_window') and self.interrupt_window:
      self.interrupt_window.destroy()
    self.intrunbut.config(state='normal')
    self.interrupt_callback(terminated)

  # ----------------------------------------------------------------------

  class ChoicePopup(Toplevel):
    def __init__(self, master, title, choices):
      Toplevel.__init__(self)
      self.title(title)
      self.master = master
      self.choices = choices
      self.buttons = []

      for choice in choices:
        but = Button(self, text = choice[0], command = choice[1])
        but.pack(side=LEFT, expand=0, fill=X, pady=3, padx=3)
        self.buttons.append(but)

  def configure_user(self, user):
    if USE_DIAMAND:
      self.user = user
      self.interaction_panel.manager.load_profile(self.user)

  def save_profile(self):
    if USE_DIAMAND:
      if not self.user:
        prompt = 'Enter a name for your profile.'
        self.user = tkSimpleDialog.askstring('Name profile',
                        prompt)
      if self.user:
        self.interaction_panel.manager.save_profile(self.user)

  def choose_profile(self, user=None):
    if user is None:
      prompt = 'Enter a name for your profile.'
      user = tkSimpleDialog.askstring('Name profile', prompt)
    if user is not None:
      self.configure_user(user)
      
  # ----------------------------------------------------------------------
  # Handling inner loops within a recorded macro
  def inner_start(self):
    self.inner_loop_start_but.configure(state='disabled')
    self.inner_loop_stop_but.configure(state='normal')
    self.add_action(None, None)
    self.vs.start_innerloop()
  def inner_iterate(self):
    self.add_action(None, None)
    self.vs.iterate_innerloop()
  def inner_done(self):
    self.inner_loop_start_but.configure(state='normal')
    self.inner_loop_stop_but.configure(state='disabled')
    self.add_action(None, None)
    self.vs.end_innerloop()
  # ----------------------------------------------------------------------

  def nextactiontype(self):
    return 'unimplemented'

# PERCY
def empty(result_set):
  if not result_set: return True
  if len(result_set.get_grouped_actions()) == 0: return True
  return False

def main():
  opts, args = getopt.getopt(sys.argv[1:], 'u:l:c:f:sd', ['line=',
    'col=', 'fontsize=', 'screenshot', 'user=', 'demo'])
  lineno = None
  colno = None
  user = None
  for (opt, val) in opts:
    if opt == '-l' or opt == '--line':
      lineno = val
    elif opt == '-u' or opt == '--user':
      user = val
      global USE_DIAMAND
      USE_DIAMAND = 1
    elif opt == '-c' or opt == '--col':
      colno = val
    elif opt == '-f' or opt == '--fontsize':
      a.default_font_size = int(val)
    elif opt == '-s' or opt == '--screenshot':
      global SCREEN_SHOT_MODE
      SCREEN_SHOT_MODE = 1
    elif opt == '-d' or opt == '--demo':
      global DEMO_MODE
      DEMO_MODE = 1
    else:
      print 'Undefined option', `opt`, 'ignored.'
  
  if DEMO_MODE:
    basefile = 'data/prob0.txt'
    title = 'SMARTedit demo'
  else:
    basefile = 'data/prob1.txt'
    title = 'SMARTedit'

  root = Tk(None, None, title)
  a = App(root)
  global app
  app = a
  if args:
    a.loadfile(args[0])
  elif os.path.exists(basefile):
    a.loadfile(basefile)
  else:
    import prob0
    a.loadcontent(prob0.PROB0)

  if lineno is not None:
    a.gotoline(lineno)
  if colno is not None:
    a.gotocol(colno)
  if user is not None:
    a.configure_user(user)
  a.mainloop()

if __name__ == '__main__':
  print 'You should be running "smartedit", not "collect.py".'

#vim:ts=5:sw=5:sts=5:
