#!/usr/bin/env python

from Tkinter import *
try:
	import Pmw
except:
	import PmwLocal as Pmw
import string, operator
import textedit
from vs import hyp_inapplicable

class ExamplePicker:
	def __init__(self, vs, state):
		self.vs = vs
		self.state = state

	def next_example(self):
		# Returns a location in the file as a potential choice for
		# jumping to.
		raise 'unimplemented abstract method; subclasses must implement'

import whrandom
class RandomExamplePicker(ExamplePicker):
	def __init__(self, vs, state):
		ExamplePicker.__init__(self, vs, state)
		self.generator = whrandom.whrandom()

	def next_example(self):
		# Returns a random location between here and the end of the file.
		# Returns None if there's nothing between here and the end of file.
		least_potential_loc = self.state.pos + 1
		potential_range = len(self.state.contents) - least_potential_loc

		# If there are no spots, return 0.
		if potential_range == 0:
			return None

		loc_choice = self.generator.randint(least_potential_loc,
									 least_potential_loc +
									 potential_range - 1)

		return '%i.%i' % \
			  textedit.rowColTuple(self.state.contents, loc_choice)

class LinewiseExamplePicker(ExamplePicker):
	def __init__(self, vs, state):
		ExamplePicker.__init__(self, vs, state)
		self.current = self.state.pos
		self.rowcol = textedit.rowColTuple(self.state.contents,
									self.current)
		

	def next_example(self):
		# Returns a position one line further along in the file.

		try:
			endline = string.index(self.state.contents, '\n',
							   self.current)
		except ValueError:
			# No more lines
			return None

		try:
			next_endline = string.index(self.state.contents, '\n',
								   endline + 1)
		except ValueError:
			next_endline = len(self.state.contents)
			
		row = self.rowcol[0] + 1
		col = min(self.rowcol[1], next_endline - endline - 1)
		new_rowcol = (row, col)

		# Update the state.
		self.rowcol = (row, self.rowcol[1])
		self.current = endline + col + 1

		return '%i.%i' % new_rowcol

class MoveBasedExamplePicker(ExamplePicker):
	DECENT_FRACTION = 0.25
	def __init__(self, vs, state):
		ExamplePicker.__init__(self, vs, state)

		# Advance past the current example
		self.next_example()

	def next_example(self):
		if not self.state:
			return None

		try:
			result_set = self.vs.execute_nth(self.state)
		except:
			return None
		
		if not result_set:
			return None

		# I'm going to try only move actions; so, pare out everything else:
		desirable = lambda hyp: hyp[0].hyp[0] == 'move' and \
				  not hyp_inapplicable(hyp[0].hyp[1])
		result_set.hypotheses = filter(desirable, result_set.hypotheses)

		if not result_set.hypotheses:
			return None

		result_set.initialize()
		(hyplist, totalprob) = result_set.pick_next_best()

		#BUG!!! Where is the (..., None) coming from?
		(command, arg, ignore) = hyplist[0][0].hyp
		assert(command == 'move')

		new_pos = textedit.positionFromTuple(self.state.contents, arg)
		predicted_pos = self.state.pos + \
					 int(self.DECENT_FRACTION *
						(new_pos - self.state.pos - 2)) + \
						1
		self.state.pos = new_pos
		
		try:
			pos = '%i.%i' % textedit.rowColTuple(self.state.contents,
									   predicted_pos)
		except:
			return None
		return pos

class ExampleChoicePopup(Toplevel):
	def __init__(self, master, num_supplied = 5):
		Toplevel.__init__(self)
		self.title('Example chooser')
		self.master = master

		# Construct a list of example generators
		vs = master.vs
		state = master.get_text_state('insert')
		self.generators = [
			RandomExamplePicker(vs, state),
			LinewiseExamplePicker(vs, state),
			MoveBasedExamplePicker(vs, state),
			]

		# Create a list of num_supplied sets of points from each generator,
		# throwing out points that come back as 'None'.
		examples = filter(None, map(lambda x: x.next_example(),
							   self.generators * num_supplied))

		# Sort and remove duplicates.
		examples.sort(lambda x, y: cmp(textedit.mark_split(x),
								 textedit.mark_split(y)))
		i = 1
		while i < len(examples):
			if examples[i] == examples[i-1]:
				del examples[i]
			else:
				i = i + 1

		# Make these the base choices
		self.choices = examples
		
		# Create widgets

		self.buttons = []

		# Manipulate existing examples
		g = Pmw.Group(self, tag_text='Manipulate examples')
		g.pack(side=TOP, expand=1, fill=BOTH)
		self.example_list = Pmw.ScrolledListBox(g.interior(),
			listbox_selectmode='extended')
		self.example_list.pack(side=LEFT, expand=1, fill=BOTH)
		h = Frame(g.interior())
		h.pack(side=RIGHT, expand=1, fill=BOTH)

		
		b = Button(h, text='Add example',
			command=self.add)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.buttons.append(b)
		b = Button(h, text='Delete example(s)',
			command=self.delete)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.buttons.append(b)
		b = Button(h, text='Go to example',
			command=self.goto)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.buttons.append(b)
		b = Button(h, text='Try this choice',
			command=self.test_choice)
		b.pack(side=TOP, expand=0, fill=NONE)
		self.buttons.append(b)

		# Button box
		buttonbox = Pmw.ButtonBox(self)
		buttonbox.pack(side=BOTTOM, expand=0, fill=X)
		buttonbox.add('OK', command=self.destroy)
		buttonbox.setdefault('OK')
		buttonbox.alignbuttons()

		self.update_examples()

	def add(self):
		location = self.master.text.index('insert')
		self.choices.append(location)
		self.update_examples()

	def delete(self):
		# Grab the selection
		selected = self.example_list.curselection()
		if not selected: return

		# Transform the selection to a list of indexes
		items = map(int, selected)

		# Reverse sort the indexes so that their deletions won't
		# interfere with each other.
		items.sort()
		items.reverse()
		
		assert(items[0] <= len(self.choices))

		# Delete all the items
		for i in items:
			del self.choices[i]

		# Reflect the changes in the example list on the screen.
		self.update_examples()
	
	def delete_all(self):
		del self.choices[:]
		self.update_examples()

	def goto(self):
		selected = self.example_list.curselection()
		if not selected: return
		items = map(int,selected)
		self.goto_list(items)

	def goto_list(self, list):
		if list:
			self.master.text.mark_set('insert', self.choices[list[0]])
			self.master.text.see('insert')
			self.master.text.focus_set()
			self.master.text.update_idletasks()		
			self.master.text.after(750,
							   lambda s=self, l=list:
							   s.goto_list(l[1:]))
	
	def test_choice(self):
		(winner, calc) = self.master.select_point(self.choices)
		#DEBUG!!!
		print 'Winner is', winner
		self.example_list.selection_clear()
		self.example_list.selection_set(winner)
		self.example_list.see(winner)

	def update_examples(self):
		"Update example list from (changed) choices"
		self.example_list.delete(0, 'end')
		for i in range(len(self.choices)):
			self.example_list.insert('end', self.choices[i])

def main():
	Popup = ScenarioPopup(None)
	Popup.mainloop()

if __name__ == '__main__':
	main()

