#!/usr/bin/env python

# VS domain code for the text editing hypothesis language

from types import *
import string, operator

def hasRow(text, row):
	# Returns whether the text buffer REALLY has the row
	# indicated. Note that this DISREGARDS the extra \n added by
	# python.

	index = 0
	search = string.index
	try:
		for k in range(row - 1):
			# May throw an exception if there's no remaining \n.
			index = search(text, '\n', index) + 1
		return 1
	except ValueError:
		return 0

def positionFromTuple(text, rowCol):
	# Throws ValueError if rowCol is not in text.
	
	index = 0
	search = string.index
	text = text + '\n'
	length = len(text)
	for k in range(rowCol[0] - 1):
		# May throw an exception if there's no remaining \n.
		index = search(text, '\n', index) + 1

	if index + rowCol[1] > length or \
	   string.find(text, '\n', index, index + rowCol[1]) != -1:
		# This tuple does not represent anything in the text buffer
		# or goes past the end of the line (so there's no position
		# corresponding to a column this far right on this row).
		raise ValueError, 'invalid (row, col) specified for text buffer'

	#HACK!!!
	# The Python text widget throws on a fake \n at the end of the file.
	# We can account for this in calculating indexes, but to avoid throwing
	# fake newlines around all over the program, we're going to finesse
	# the issue here and move the selection back up one character. This
	# could cause some problems (e.g., if the user cuts/copies and then
	# pastes, that will be different depending on the \n) but should be
	# fine if the user is just selecting and then deleting the selection.
	if index + rowCol[1] == length:
		index = index - 1
		
	return index + rowCol[1]


def rowColTuple(text, loc):
	"""Given the text buffer and a location as an integer character
	location in that buffer, returns a tuple (row, col) in the buffer."""

	if loc == len(text):
		# This is the final row/column in the text. Works fine below.
		pass
	elif loc > len(text):
		raise ValueError, 'invalid position in text buffer in rowColTuple'

	# Because rows are numbered from 1 and to handle the OB1 and
	# boundary case for column search, prepend a '\n'.
	text = '\n' + text
	loc = loc + 1
	return (string.count(text, '\n', 0, loc),
		   loc - string.rfind(text, '\n', 0, loc) - 1)

# ----------------------------------------------------------------------
# This stuff is hypothesis language-dependent

def mark_split(mark):
	return map(int, string.split(mark, '.'))

def mark_add(p1, p2):
	line1, char1 = mark_split(p1)
	line2, char2 = mark_split(p2)
	return '%i.%i' % (line1+line2, char1+char2)

# ---------------------------------------------------------------------------
# Interpret a string representation of a list of events
def interpret(event_str):
	actions = filter(None, string.split(event_str, '\n'))
	return map(eval, actions)

# ---------------------------------------------------------------------------
# Abstract event types

CLIPBOARD = None

# Abstract base class
class Abstract:
	def __init__(self):
		self.type = 'abstract'
	
	def __cmp__(self, newevent):
		return cmp(self.__dict__, newevent.__dict__)
	
	def __str__(self):
		return '%s()' % self.type
	__repr__ = __str__

class Insert(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'insert'
		# self.text is either a string, or None (bottom)
		self.text = outstate.contents[instate.pos:outstate.pos]
		self.instate = instate
		self.outstate = outstate

	def __str__(self):
		return '%s(text=%s)' % (self.type, `self.text`)
	__repr__ = __str__

class Select(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'select'
		self.instate = instate
		self.outstate = outstate

	def __str__(self):
		return '%s(%s, %s)' % (self.type, self.instate.pos,
						   self.outstate.pos)
	__repr__ = __str__

class Delete(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'delete'
		self.instate = instate
		self.outstate = outstate
	
	def __str__(self):
		return '%s(%s, %s)' % (self.type, self.instate.pos,
						   self.outstate.pos)
		return '%s(Location)' % self.type
	__repr__ = __str__

class Move(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)

		self.type = 'move'
		self.instate = instate
		self.outstate = outstate
	
	def __str__(self):
		return '%s(%s, %s)' % (self.type, self.instate.pos,
						   self.outstate.pos)
	__repr__ = __str__

class Copy(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'copy'
		self.instate = instate
		self.outstate = outstate
	
class Paste(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'paste'
		self.instate = instate
		self.outstate = outstate
	
class Cut(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'cut'
		self.instate = instate
		self.outstate = outstate

class DeleteSelection(Abstract):
	def __init__(self, instate, outstate):
		Abstract.__init__(self)
		self.type = 'deleteselection'
		self.instate = instate
		self.outstate = outstate
	
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
