class Trie:
    def __init__(self):
        self.NUM_LETTERS = 128

        # -1 == unseen
        # 0 == negative
        # 1 or greater == number of postive examples
        self.data = -1 # initialize data to unseen

        # this is the backup field.  save() stores
        #   the current value of data in it, and restore()
        #   stores this value into data
        self.backup = -1

        # links to the letters
        self.links = [None] * self.NUM_LETTERS

        # link to this node's parent
        self.parent = None

    # text is the text to be inserted
    # value is int w/ 1 == pos, 0 == neg, -1 == unseen
    def insert(self, text, value):
        ptr = self
        for i in text:
            index = ord(i)
            if ptr.links[index] == None:
                ptr.links[index] = Trie()
                ptr.links[index].parent = ptr
            ptr = ptr.links[index]

            if value == 0:
                ptr.data = 0
            elif value == 1:
                if ptr.data != 0:
                    if ptr.data == -1:
                        ptr.data = 1
                    else:
                        ptr.data = ptr.data + 1

    def save(self):
        self.backup = self.data
        
        i = 0
        while i < self.NUM_LETTERS:#len(self.links):
            if self.links[i] != None:
                self.links[i].save()
            i = i + 1

    def restore(self):
        self.data = self.backup
        
        i = 0
        while i < self.NUM_LETTERS:#len(self.links):
            if self.links[i] != None:
                self.links[i].restore()
            i = i + 1

    def search(self, text):
        ptr = self
        for i in text:
            index = ord(i)
            ptr = ptr.links[index]
            if ptr == None or \
               (ptr.data == -1 and ptr.parent != None):
                return -1
        return ptr.data

    def seen(self):
        if self.data < -1:
            return 0
        return 1

    def status(self):
        return self.data

    def output(self, depth):
        i = 0
        while i < self.NUM_LETTERS:#len(self.links):
            if self.links[i] != None:
                print '%s%s %i' % ('  ' * depth, chr(i), self.links[i].data)
                self.links[i].output(depth+1)
            i = i + 1
