package programinduction.arith

import scala.collection.mutable.{ArrayBuffer,HashMap,HashSet}
import fig.exec.Execution
import tea.Utils
import tea.Utils.{map,foreach,fmt,fmts,fmt1,fmt2,returnFirst,assertValid,impossible,fails}
import tea.Utils.{track,begin_track,end_track,logs,dbgs}

import java.util.{Set,ArrayList,Random}
import scala.reflect.Manifest

import programinduction._
import StandardLibrary._
import Router.{B,C,S}
import programinduction.{MPrimitiveNode => __}
import programinduction.{MCompositeNode => ___}
import MyUtils._

case class IntPair(first:Int, second:Int)

case class ArithmeticWorld(worldOpts:WorldOptions, arithOpts:ArithOptions) extends AbstractWorld {
  trait TestCase extends AbstractTestCase
  case class TestCase1(i:Int, o:Int) extends TestCase {
    val inputs = WrappedValue(i) :: Nil
    val output = WrappedValue(o)
  }
  case class TestCase2(i1:Int, i2:Int, o:Int) extends TestCase {
    val inputs = WrappedValue(i1) :: WrappedValue(i2) :: Nil
    val output = WrappedValue(o)
  }
  case class TestCase3(i1:Int, i2:Int, i3:Int, o:Int) extends TestCase {
    val inputs = WrappedValue(i1) :: WrappedValue(i2) :: WrappedValue(i3) :: Nil
    val output = WrappedValue(o)
  }
  
  import ArithmeticLibrary._

  def allowCombination_?(func:KNode, arg:KNode) : Boolean = {
    (func, arg) match {
      // Don't allow noops
      case (KPrimitiveNode(Add), KPrimitiveNode(WrappedValue(0))) => false
      case (KCompositeNode(_, Add, _), KPrimitiveNode(WrappedValue(0))) => false
      case (KCompositeNode(_, Sub, _), KPrimitiveNode(WrappedValue(0))) => false
      case _ => true
    }
  }

  def computeRogueProb(c:Combinator) = 1.0

  def populatePrimitives = {
    val primitives = new ArrayBuffer[Combinator]

    //primitives ++= List(Add, Sub, WrappedValue(1))

    primitives ++= List(Lt, Gt, Eq)
    primitives ++= List(If(IntType))
    primitives ++= List(Add, Sub, Mul, Div)
    primitives ++= (arithOpts.intPrimitiveRange.first to arithOpts.intPrimitiveRange.second).map(WrappedValue(_))

    primitives.foreach(addPrimitive)
  }

  def populateScenarios = {
    def A(func:MNode, arg:MNode) = ___(Nil, func, arg)
    def V(value:Combinator) = __(value)
    arithOpts.testLabel match {
      case "basic" => // Basic
        scenarios += new Scenario("+1", Array(TestCase1(1, 2), TestCase1(2, 3))) {
          //override def getRoots = ___(Nil, __(Add), ___(Nil, ___(Nil, __(Sub), __(2)), __(1))) :: Nil
        }
      case "sharing" => // Test sharing with provided roots
        scenarios += new Scenario("+1", Array(TestCase1(1, 2), TestCase1(2, 3))) {
          override def getRoots = ___(Nil, __(Add), ___(Nil, ___(Nil, __(Sub), __(2)), __(1))) :: Nil
        }
        scenarios += new Scenario("+4", Array(TestCase1(1, 5))) {
          override def getRoots = ___(Nil, __(Add), ___(Nil, ___(Nil, __(Sub), __(5)), __(1))) :: Nil
        }
      case "root-switching" => // Test root switching (don't set changeCurrRootProb too high or else won't mix)
        // Root switching is wrong!
        scenarios += new Scenario("+1", Array(TestCase1(1, 2), TestCase1(2, 3))) {
          val candidates = ___(Nil, __(Add), ___(Nil, ___(Nil, __(Sub), __(2)), __(1))) ::
                           ___(Nil, __(Add), ___(Nil, ___(Nil, __(Sub), __(3)), __(2))) ::
                           Nil
          addEDenotationCandidates(testCases.toList, candidates)
          override def getRoots = candidates.head :: Nil
        }
      case "test" =>
        //scenarios += Scenario("max0", Array(TestCase1(8, 8), TestCase1(-4, 0))) // keep non-negative
        scenarios += new Scenario("test", Array(TestCase2(1, 2, 3))) {
          override def getRoots = __(Add) :: Nil
        }
      case "motivation" =>
        // This example illustrates the benefit of sharing.
        // There are four settings.  logProbs at temperature = 4
        // We want to initialize in the upper-left and end up into the lower-right.
        //                        +,-,*,/           min,max
        // non-refactored         -19.7             -18.55
        // refactored             -15.19            -13.016
        // Get good result for temperature = 10

        def cmpNode(func:Combinator) = {
          if (arithOpts.initWithRefactoredVersion) {
            // (SS (CBC (BC if I) I) <) [refactored version] for comparison function - hopefully we'll end up here
            ___(S::S::Nil, ___(C::B::C::Nil, ___(B::C::Nil, __(If[Int]), __(Id[Int])), __(Id[Int])), __(func))
          }
          else {
            // (CS (SC (BB if <) I) I) [initial version] for comparison function
            ___(C::S::Nil, ___(S::C::Nil, ___(B::B::Nil, __(If[Int]), __(func)), __(Id[Int])), __(Id[Int]))
          }
        }

        scenarios += new MultiCandidateScenario("min", Array(TestCase2(3, 2, 2), TestCase2(7, 4, 4))) {
          def candidates = ___(C::C::Nil, ___(B::B::Nil, __(Add), __(Sub)), __(1)) :: // (a - b + 1)
                           cmpNode(Lt) :: // min
                           Nil
        }

        scenarios += new MultiCandidateScenario("max", Array(TestCase2(3, 6, 6), TestCase2(2, 4, 4))) {
          def candidates = ___(B::C::Nil, ___(B::Nil, __(Add), ___(C::Nil, __(Div), __(2))), __(Id[Int])) :: // (b/2 + a)
                           cmpNode(Gt) :: // max
                           Nil
        }

        // Complex
        /*scenarios += new MultiCandidateScenario("max", Array(TestCase2(2, 1, 2), TestCase2(1, 2, 2))) {
          def ab_div2 = ___(C::C::Nil, ___(B::B::Nil, __(Div), __(Mul)), __(2))
          def candidates = ___(C::C::Nil, ___(B::B::Nil, __(Add), ab_div2), __(1)) :: // (ab/2 + 1)
                           cmpNode(Gt) :: // max
                           Nil
        }*/

        // This is too simple
        /*scenarios += new MultiCandidateScenario("max", Array(TestCase2(3, 2, 3), TestCase2(2, 3, 3))) {
          def candidates = ___(C::C::Nil, ___(B::B::Nil, __(Sub), __(Mul)), __(3)) :: // (ab - 3)
                           cmpNode(Gt) :: // max
                           Nil
        }*/
        // This is too complicated to show in the paper
        /*scenarios += new MultiCandidateScenario("minmax", Array(TestCase3(2, 6, 4, 4), TestCase3(2, 3, 4, 3))) { // min(max(a,b), c)
          def candidates =
            ___(C::C::C::Nil, ___(B::B::B::Nil, __(Div), ___(B::B::Nil, __(Add), __(Add))), __(3)) :: // (a+b+c)/3
            ___(C::C::B::Nil, ___(B::B::Nil, cmpNode(Lt), cmpNode(Gt)), __(Id[Int])) :: // minmax
            Nil
        }*/

      case "fulltest" =>
        scenarios += Scenario("1[x<y]", Array(TestCase2(5, 2, 0), TestCase2(4, 7, 1))) // less than indicator
        scenarios += Scenario("1[x>y]", Array(TestCase2(5, 2, 1), TestCase2(4, 7, 0))) // greater than indicator
        scenarios += Scenario("1[x=y]", Array(TestCase2(5, 2, 0), TestCase2(4, 4, 1))) // equal indicator
        scenarios += Scenario("min", Array(TestCase2(5, 2, 2), TestCase2(4, 7, 4))) // min
        scenarios += Scenario("max", Array(TestCase2(5, 2, 5), TestCase2(4, 7, 7))) // max
        scenarios += Scenario("abs", Array(TestCase1(-3, 3), TestCase1(4, 4))) // absolute value
        scenarios += Scenario("max0", Array(TestCase1(8, 8), TestCase1(-4, 0))) // keep non-negative
      case "fulltest-auto" =>
        scenarios += makeScenario("1[x<y]", { (x:Int,y:Int) => if (x<y) 1 else 0 })
        scenarios += makeScenario("1[x>y]", { (x:Int,y:Int) => if (x>y) 1 else 0 })
        scenarios += makeScenario("1[x=y]", { (x:Int,y:Int) => if (x==y) 1 else 0 })
        scenarios += makeScenario("min", { (x:Int,y:Int) => if (x<y) x else y })
        scenarios += makeScenario("max", { (x:Int,y:Int) => if (x>y) x else y })
        scenarios += makeScenario("abs", { x:Int => x.abs })
        scenarios += makeScenario("max0", { x:Int => x max 0 })
      case _ => throw fails("Invalid testLabel: %s", arithOpts.testLabel)
    }
  }

  // Automatically generate random test cases
  def rand = worldOpts.testCaseRandom.nextInt(11)-5
  def makeScenario(name:String, f:Int=>Int) = {
    Scenario(name, map(worldOpts.numTestCases, { i:Int =>
      val x = rand
      TestCase1(x, f(x))
    }))
  }
  def makeScenario(name:String, f:(Int,Int)=>Int) = {
    Scenario(name, map(worldOpts.numTestCases, { i:Int =>
      val (x, y) = (rand, rand)
      TestCase2(x, y, f(x, y))
    }))
  }

  case class Scenario(name:String, testCases:Seq[TestCase]) extends AbstractScenario {
    def getRoots = exhaustiveSearch.map(MNode.fromKNode)
    def createInitRootNodes = {
      roots ++= getRoots
    }
  }

  abstract class MultiCandidateScenario(name:String, testCases:Seq[TestCase]) extends Scenario(name, testCases) {
    def candidates : Seq[MNode]
    addEDenotationCandidates(trainTestCases.toList, candidates)
    override def getRoots = {
      (if (arithOpts.initWithLastCandidate) candidates.last else candidates.head) :: Nil
    }
  }
}

class ArithOptions {
  import fig.basic.Option
  @Option var testLabel = ""
  @Option var intPrimitiveRange = new fig.basic.IntPair(0, 1)

  @Option var initWithLastCandidate = false
  @Option var initWithRefactoredVersion = false
}

class ArithmeticTestRunner extends Runnable {
  import fig.basic.OptionSet
  @OptionSet(name="world") val worldOpts = new WorldOptions
  @OptionSet(name="arith") val arithOpts = new ArithOptions

  def run = {
    val world = new ArithmeticWorld(worldOpts, arithOpts)
    world.initialize
    world.learn
  }
}

object ArithmeticTest {
  def main(args:Array[String]) = Execution.run(args, "main", new ArithmeticTestRunner)
}
